import re
import pathlib
import numpy as np
import skimage.io
from . import reg
from .fileseries import _read


# Classes for reading datasets consisting of TIFF files with a naming pattern.
# The pattern must include row and column as integers, and optionally a channel
# name or number.
#
# This code is experimental and probably still has a lot of rough edges.


class FilePatternMetadata(reg.Metadata):

    def __init__(self, path, pattern, overlap, pixel_size):
        # The pattern argument uses the Python Format String syntax with
        # required "row" and "col" and optionally "channel" fields. A width
        # specification with leading zeros must be used for any fields that are
        # zero-padded. The pattern is used both to parse the filenames upon
        # initialization as well as to synthesize filenames when reading images.
        # Example pattern: 'img_{channel}_r{row:03}_c{col:03}.tif'
        self.path = pathlib.Path(path)
        self.pattern = pattern
        self.overlap = overlap
        self._pixel_size = pixel_size
        self._enumerate_tiles()

    def _enumerate_tiles(self):
        # Translate a restricted subset of the "format" pattern language to
        # a matching regex with named capture.
        pattern = self.pattern.replace('.', '\.')
        pattern = pattern.replace('(', '\(')
        pattern = pattern.replace(')', '\)')
        regex = re.sub(r'{([^:}]+)(?:[^}]*)}', r'(?P<\1>.*?)', pattern)
        rows = set()
        cols = set()
        channels = set()
        n = 0
        for p in self.path.iterdir():
            match = re.match(regex, p.name)
            if match:
                gd = match.groupdict()
                rows.add(int(gd['row']))
                cols.add(int(gd['col']))
                channels.add(gd.get('channel'))
                n += 1
        if n != len(rows) * len(cols) * len(channels):
            raise Exception("Tiles do not form a full rectangular grid")
        self._actual_num_images = len(rows) * len(cols)
        self.channel_map = dict(enumerate(sorted(channels)))
        self.height = len(rows)
        self.width = len(cols)
        self.row_offset = min(rows)
        self.col_offset = min(cols)
        path = self.path / self.pattern.format(
            row=self.row_offset, col=self.col_offset,
            channel=self.channel_map[0]
        )
        img = _read(path)
        if img.ndim not in (2, 3):
            raise Exception(f"Image must have 2 or 3 dimensions: {path}")
        self._tile_size = np.array(img.shape[-2:])
        self._dtype = img.dtype
        self.multi_channel_tiles = False
        # Handle multi-channel tiles (pattern must not include channel).
        if len(self.channel_map) == 1 and img.ndim == 3:
            self.channel_map = {c: None for c in range(img.shape[0])}
            self.multi_channel_tiles = True
        self._num_channels = len(self.channel_map)

    @property
    def _num_images(self):
        return self._actual_num_images

    @property
    def num_channels(self):
        return self._num_channels

    @property
    def pixel_size(self):
        return self._pixel_size

    @property
    def pixel_dtype(self):
        return self._dtype

    def tile_position(self, i):
        row, col = self.tile_rc(i)
        return [row, col] * self.tile_size(i) * (1 - self.overlap)

    def tile_size(self, i):
        return self._tile_size

    def tile_rc(self, i):
        row = i // self.width + self.row_offset
        col = i % self.width + self.col_offset
        return row, col


class FilePatternReader(reg.Reader):

    def __init__(self, path, pattern, overlap, pixel_size=1.0):
        # See FilePatternMetadata for an explanation of the pattern syntax.
        self.path = pathlib.Path(path)
        self.pattern = pattern
        self.overlap = overlap
        self.metadata = FilePatternMetadata(
            self.path, self.pattern, overlap, pixel_size
        )

    def read(self, series, c):
        path = self.path / self.filename(series, c)
        channel = c if self.metadata.multi_channel_tiles else 0
        img = _read(path, channel)
        return img

    def filename(self, series, c):
        row, col = self.metadata.tile_rc(series)
        c = self.metadata.channel_map[c]
        return self.pattern.format(row=row, col=col, channel=c)



import warnings
import sys
import argparse
import numpy as np
import skimage.transform
import skimage.util
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import matplotlib.text as mtext
from .. import reg, utils
from ..scripts import ashlar as ascript


def main(argv=sys.argv):

    parser = argparse.ArgumentParser(
        description=(
            "Stitch a multi-tile image using the nominal stage positions and"
            " display the resulting image."
        )
    )
    parser.add_argument(
        "input", help="Path to image (BioFormats supported formats only)"
    )
    parser.add_argument(
        "-n", "--numbers", action="store_true", help="Display tile numbers"
    )
    parser.add_argument(
        "-b", "--bounds", action="store_true", help="Display tile bounds"
    )
    parser.add_argument(
        "-c", "--channel", type=int, default=0,
        help="Channel number to display; default: 0",
    )
    parser.add_argument(
        "-d", "--downsample", metavar="FACTOR", type=float, default=10,
        help="Downsample the image resolution by this factor for display (reduces"
        " memory requirements and improves performance); default: 10",
    )
    parser.add_argument(
        "-l", "--log", action="store_true", help="Log-transform pixel intensities"
        " (helps visualize dim images)",
    )
    parser.add_argument(
        '--flip-x', default=False, action='store_true',
        help='Flip tile positions left-to-right',
    )
    parser.add_argument(
        '--flip-y', default=False, action='store_true',
        help='Flip tile positions top-to-bottom',
    )
    args = parser.parse_args()

    reader = ascript.build_reader(args.input)
    ascript.process_axis_flip(reader, args.flip_x, args.flip_y)
    metadata = reader.metadata

    resolution_scale = 1 / args.downsample
    positions = (metadata.positions - metadata.origin) * resolution_scale
    pmax = (positions + metadata.size * resolution_scale).max(axis=0)
    mshape = (pmax + 0.5).astype(int)
    mosaic = np.zeros(mshape, dtype=np.uint16)

    total = reader.metadata.num_images
    for i in range(total):
        sys.stdout.write("\rLoading %d/%d" % (i + 1, total))
        sys.stdout.flush()
        img = reader.read(c=args.channel, series=i)
        img = skimage.transform.rescale(img, resolution_scale, anti_aliasing=False)
        img = skimage.img_as_uint(img)
        if args.log:
            intensity_scale = 65535 / np.log(65535)
            img = (np.log(np.maximum(img, 1)) * intensity_scale).astype(np.uint16)
        # Round position so paste will skip the expensive subpixel shift.
        pos = np.round(positions[i])
        utils.paste(mosaic, img, pos, np.maximum)
    print()

    ax = plt.gca()

    plt.imshow(X=mosaic, axes=ax, extent=(0, pmax[1], pmax[0], 0))

    h, w = metadata.size * resolution_scale
    for i, (x, y) in enumerate(np.fliplr(positions)):
        if args.bounds:
            rect = mpatches.Rectangle((x, y), w, h, color='black', fill=False)
            ax.add_patch(rect)
        if args.numbers:
            xc = x + w / 2
            yc = y + h / 2
            circle = mpatches.Circle((xc, yc), w / 5, color='salmon', alpha=0.5)
            text = mtext.Text(
                xc, yc, str(i), color='k', size=10, ha='center', va='center'
            )
            ax.add_patch(circle)
            ax.add_artist(text)

    plt.show()


if __name__ == '__main__':
    main()



import numpy as np
from skimage.transform import warp


def _barrel_mapping(xy, center, k):
    x, y = xy.T.astype(float)
    y0, x0 = center
    x -= x0
    y -= y0
    # The warp mapping function defines the INVERSE map to apply, which in the
    # case of a distortion correction is the FORWARD map of the distortion
    # itself. See Fitzgibbon 2001 eq 1 for details.
    r2 = x ** 2 + y ** 2
    f = 1 + k * r2
    xy[..., 0] = x * f + x0
    xy[..., 1] = y * f + y0
    return xy


def barrel_correction(
    image,
    k,
    center=None,
    output_shape=None,
    order=1,
    mode=None,
    cval=0,
    clip=True,
    preserve_range=False,
):
    """Perform a transform to correct for barrel or pincushion distortion.

    Parameters
    ----------
    image : ndarray
        Input image.
    k : float
        Distortion parameter
    center : (row, column) tuple or (2,) ndarray, optional
        Center coordinate of transformation.

    Returns
    -------
    cartesian : ndarray
        Cartesian version of the input.
        Rows correspond to radius and columns to angle values.

    Other parameters
    ----------------
    output_shape : tuple (rows, cols), optional
        Shape of the output image generated. By default the shape of the input
        image is preserved.
    order : int, optional
        The order of the spline interpolation, default is 1. The order has to
        be in the range 0-5. See `skimage.transform.warp` for detail.
    mode : {'constant', 'edge', 'symmetric', 'reflect', 'wrap'}, optional
        Points outside the boundaries of the input are filled according
        to the given mode, with 'constant' used as the default. Modes match
        the behaviour of `numpy.pad`.
    cval : float, optional
        Used in conjunction with mode 'constant', the value outside
        the image boundaries.
    clip : bool, optional
        Whether to clip the output to the range of values of the input image.
        This is enabled by default, since higher order interpolation may
        produce values outside the given input range.
    preserve_range : bool, optional
        Whether to keep the original range of values. Otherwise, the input
        image is converted according to the conventions of `img_as_float`.

    Notes
    -----
    Radial distortion is modeled here using a one-parameter model described in
    [1]_ (Equation 1).

    References
    ----------
    .. [1] A. W. Fitzgibbon, "Simultaneous linear estimation of multiple view
       geometry and lens distortion," Proceedings of the 2001 IEEE Computer
       Society Conference on Computer Vision and Pattern Recognition. CVPR 2001,
       Kauai, HI, USA, 2001, pp. I-I. :DOI:`10.1109/CVPR.2001.990465`.

    """

    if mode is None:
        mode = "constant"

    if center is None:
        center = np.array(image.shape)[:2] / 2

    warp_args = {"center": center, "k": k}

    return warp(image, _barrel_mapping, map_args=warp_args,
                output_shape=output_shape, order=order, mode=mode, cval=cval,
                clip=clip, preserve_range=preserve_range)



import functools
import itertools
import warnings
import skimage
import scipy.ndimage
import numpy as np


# Pre-calculate the Laplacian operator kernel. We'll always be using 2D images.
_laplace_kernel = skimage.restoration.uft.laplacian(2, (3, 3))[1]

def whiten(img, sigma):
    img = skimage.img_as_float32(img)
    if sigma == 0:
        output = scipy.ndimage.convolve(img, _laplace_kernel)
    else:
        output = scipy.ndimage.gaussian_laplace(img, sigma)
    return output


@functools.lru_cache
def get_window(shape):
    # Build a 2D Hann window by taking the outer product of two 1-D windows.
    wy = np.hanning(shape[0]).astype(np.float32)
    wx = np.hanning(shape[1]).astype(np.float32)
    window = np.outer(wy, wx)
    return window


def window(img):
    assert img.ndim == 2
    return img * get_window(img.shape)


def register(img1, img2, sigma, upsample=10):
    img1w = window(whiten(img1, sigma))
    img2w = window(whiten(img2, sigma))
    shift = skimage.registration.phase_cross_correlation(
        img1w,
        img2w,
        upsample_factor=upsample,
        normalization=None,
        return_error=False,
    )
    # At this point we may have a shift in the wrong quadrant since the FFT
    # assumes the signal is periodic. We test all four possibilities and return
    # the shift that gives the highest direct correlation (sum of products).
    shape = np.array(img1.shape)
    shift_pos = (shift + shape) % shape
    shift_neg = shift_pos - shape
    shifts = list(itertools.product(*zip(shift_pos, shift_neg)))
    correlations = [
        np.abs(np.sum(img1w * scipy.ndimage.shift(img2w, s, order=0)))
        for s in shifts
    ]
    idx = np.argmax(correlations)
    shift = shifts[idx]
    correlation = correlations[idx]
    total_amplitude = np.linalg.norm(img1w) * np.linalg.norm(img2w)
    if correlation > 0 and total_amplitude > 0:
        error = -np.log(correlation / total_amplitude)
    else:
        error = np.inf
    return shift, error


def nccw(img1, img2, sigma):
    img1w = whiten(img1, sigma)
    img2w = whiten(img2, sigma)
    correlation = np.abs(np.sum(img1w * img2w))
    total_amplitude = np.linalg.norm(img1w) * np.linalg.norm(img2w)
    if correlation > 0 and total_amplitude > 0:
        diff = correlation - total_amplitude
        if diff <= 0:
            error = -np.log(correlation / total_amplitude)
        elif diff < 2e-5:
            # This situation can occur due to numerical precision issues when
            # img1 and img2 are very nearly or exactly identical. If the
            # difference is small enough, let it slide.
            error = 0
        else:
            raise RuntimeError(
                f"correlation > total_amplitude (diff={diff})"
            )
    else:
        error = np.inf
    return error


def crop(img, offset, shape):
    # Note that this only crops to the nearest whole-pixel offset.
    start = offset.round().astype(int)
    end = start + shape
    img = img[start[0]:end[0], start[1]:end[1]]
    return img


# TODO:
# - Deal with ringing from high-frequency elements. The wrapped edges of the
#   image are especially bad, where the wrapping introduces sharp
#   discontinuities. The edge artifacts could be dealt with in several ways
#   (extend the trailing image edge via mirroring, throw away some of the
#   trailing edge of the shifted result) but edges in the "true" image content
#   would require proper pre-filtering. What filter to use, and how to apply it
#   quickly?
# - Can we use real FFT for a ~50% overall speedup? Fourier-space matrices will
#   all be half-size in the last dimension, so FFT is around 50% faster and our
#   fshift calculations will be too.
# - Trailing edge pixels should be zeroed to match the behavior of
#   scipy.ndimage.shift, which we rely on in our maximum-intensity projection.
def fourier_shift(img, shift):
    # Ensure properly aligned complex64 data (fft requires complex to avoid
    # reallocation and copying).
    img = skimage.util.img_as_float32(img)
    img = pyfftw.byte_align(img, dtype=np.complex64)
    # Compute per-axis frequency values according to the Fourier shift theorem.
    # (Read "w" here as "omega".) We pre-multiply as many scalar values as
    # possible on these vectors to avoid operations on the full w matrix below.
    v = np.fft.fftfreq(img.shape[0])
    wy = (2 * np.pi * v * shift[0]).astype(np.float32).reshape(-1, 1)
    u = np.fft.fftfreq(img.shape[1])
    wx = (2 * np.pi * u * shift[1]).astype(np.float32)
    # Add column and row vector to get full expanded matrix of frequencies.
    w = wy + wx
    # We perform an explicit application of Euler's formula with careful
    # management of output arrays to avoid extra memory allocations and copies,
    # squeezing out some speed over the obvious np.exp(-1j*w).
    fshift = np.empty_like(img, dtype=np.complex64)
    np.cos(w, out=fshift.real)
    np.sin(w, out=fshift.imag)
    np.negative(fshift.imag, out=fshift.imag)
    # Perform the FFT, multiply in-place by the shift matrix, then IFFT.
    freq = pyfftw.builders.fft2(img, planner_effort='FFTW_ESTIMATE',
                                avoid_copy=True, auto_align_input=True,
                                auto_contiguous=True)()
    freq *= fshift
    img_s = pyfftw.builders.ifft2(freq, planner_effort='FFTW_ESTIMATE',
                                  avoid_copy=True, auto_align_input=True,
                                  auto_contiguous=True)()
    # Any non-zero imaginary component of the resulting array is due to
    # numerical error, so we can just return the real part.
    # FIXME need to zero out row(s) and column(s) we shifted away from,
    # since at this point we have a cyclic rotation rather than a shift.
    return img_s.real


def paste(target, img, pos, func=None):
    """Composite img into target. from ASHLAR"""
    pos = np.array(pos)
    # Bail out if destination region is out of bounds.
    if np.any(pos >= target.shape[:2]) or np.any(pos + img.shape[:2] < 0):
        return
    pos_f, pos_i = np.modf(pos)
    yi, xi = pos_i.astype('i8')
    # Clip img to the edges of the mosaic.
    if yi < 0:
        img = img[-yi:]
        yi = 0
    if xi < 0:
        img = img[:, -xi:]
        xi = 0
    target_slice = target[yi:yi+img.shape[0], xi:xi+img.shape[1]]
    
    def crop_like(img, target):
        if (img.shape[0] > target.shape[0]):
            img = img[:target.shape[0], :]
        if (img.shape[1] > target.shape[1]):
            img = img[:, :target.shape[1]]
        return img
    
    img = crop_like(img, target_slice)
    # Skip expensive sub-pixel shift if fractional position is zero.
    if pos_f.any():
        if img.ndim == 2:
            img = scipy.ndimage.shift(img, pos_f)
        else:
            for c in range(img.shape[2]):
                img[...,c] = scipy.ndimage.shift(img[...,c], pos_f)
        # For any axis where there is a non-zero subpixel shift, crop out the
        # last row or column of pixels on the "losing" side. These pixels will
        # be darker than normal and will introduce artifacts in most blending
        # modes.
        y1 = None if pos_f[0] <= 0 else 1
        y2 = None if pos_f[0] >= 0 else -1
        x1 = None if pos_f[1] <= 0 else 1
        x2 = None if pos_f[1] >= 0 else -1
        img = img[y1:y2, x1:x2]
        target_slice = target_slice[y1:y2, x1:x2]
        # Exit if image area is zero after subpixel shift.
        if not np.all(img.shape):
            return
    if np.issubdtype(img.dtype, np.floating):
        np.clip(img, 0, 1, img)
    img = dtype_convert(img, target.dtype)
    if func is None:
        target_slice[:] = img
    elif isinstance(func, np.ufunc):
        func(target_slice, img, out=target_slice)
    else:
        target_slice[:] = func(target_slice, img)


def pastefunc_blend(target, img):
    """Linear blend based on distance to unfilled space in target."""
    # This should catch actual holes but not the actual unfilled space.
    # FIXME Should generate mask from tile boundaries instead.
    hole_threshold = np.mean(target.shape)
    mask = skimage.morphology.remove_small_holes(target != 0, hole_threshold)
    dist = scipy.ndimage.distance_transform_cdt(mask)
    dmax = dist.max()
    if dmax == 0:
        alpha = 0
    else:
        alpha = dist / dmax
        # Keep target pixel values where img has value 0 (with a 1-pixel
        # dilation to clean up the edge). This is a temporary hack to support
        # image corrections that leave regions of zero pixels around the edge of
        # img, such as barrel correction and rotation.
        # FIXME Should compute the geometry of the source image mask more
        # deliberately and precisely.
        alpha[skimage.morphology.binary_dilation(img == 0)] = 1
    return target * alpha + img * (1 - alpha)


def crop_like(img, target):
    if (img.shape[0] > target.shape[0]):
        img = img[:target.shape[0], :]
    if (img.shape[1] > target.shape[1]):
        img = img[:, :target.shape[1]]
    return img


def dtype_convert(img, dtype):
    """Convert an image to the requested data-type.

    This is just a wrapper around skimage.util.dtype.convert that silences its
    FutureWarning, as Ashlar pins skimage to a version before that planned
    deprecation.

    """
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", r".*scikit-image 1\.0", FutureWarning)
        return skimage.util.dtype.convert(img, dtype)


def imsave(fname, arr, **kwargs):
    """Save an image to file.

    This is a wrapper around skimage.io.imsave to force check_contrast=False
    since the contrast check kills us on the huge images we create.

    """

    if "check_contrast" in kwargs:
        warnings.warn("ignoring check_contrast argument -- forcing to False")
    kwargs["check_contrast"] = False

    # We use scikit-image's vendored copy of tifffile directly rather than allow
    # scikit-image to optimistically use a separately-installed copy of tifffile
    # due to bugs and API inconsistencies in the latest pypi-hosted version:
    # * Use of "centimeter" for resolution units instead of "cm"
    # * A bug in writing single-tile planes -- issue #3 on GitHub
    # FIXME Once scikit-image un-vendors tifffile (#4235) AND tifffile fixes #3
    # we can remove this block and use `skimage.io.imsave` directly again. Or we
    # might just want to switch to tifffile.imsave.
    del kwargs["check_contrast"]
    import skimage.external.tifffile
    skimage.external.tifffile.imsave(fname, arr, **kwargs)



import warnings
import sys
import glob
import re
import subprocess
import gc
import numpy as np
import skimage.transform
import skimage.exposure
from .. import utils


# Target width.
TW = 1920

FORMAT = "frame_%04d.jpg"
MOVIE_FILENAME = 'alignment.mp4'

CMD = ("ffmpeg -v error -r 5 -y -i " + FORMAT + " -an"
       " -vcodec libx264 -profile:v main -level 3 -pix_fmt yuv420p -crf 18"
       " " + MOVIE_FILENAME)


def main(argv=sys.argv):
    # Sort paths by cycle number, numerically.
    paths = sorted(glob.glob('cycle_*_channel_0.tif'),
                   key=lambda s: int(s.split('_')[1]))
    for i, in_path in enumerate(paths):
        out_path = FORMAT % i
        print('%s -> %s' % (in_path, out_path))
        img = skimage.io.imread(in_path)
        if i == 0:
            h, w = img.shape
            # Tall images will be rotated by 90 degrees.
            rotate = h > w
            if rotate:
                w, h = h, w
            scale = w / TW
            # Ensure final dimensions are multiples of 2, per video codec
            # restrictions.
            w = int(w / scale // 2 * 2)
            h = int(h / scale // 2 * 2)
        if rotate:
            img = np.rot90(img)
        img_new = skimage.transform.resize(img, (h, w), mode='reflect')
        # Free this memory as soon as possible.
        del img
        gc.collect()
        bins = np.linspace(0, img_new.max(), 3000)
        counts, _ = np.histogram(img_new, bins=bins)
        # Find peak, skipping first and last bin.
        vmin = bins[np.argmax(counts[1:-1]) + 1]
        vmax = np.percentile(img_new, 99.5)
        img_new = skimage.exposure.rescale_intensity(img_new, (vmin, vmax))
        img_new = skimage.exposure.adjust_gamma(img_new, 1/2.2)
        utils.imsave(out_path, img_new)
    print('rendering frames to %s' % (MOVIE_FILENAME))
    subprocess.call(CMD.split(' '))
    return 0


if __name__ == '__main__':
    main()



import sys
import math
import warnings
import xml.etree.ElementTree
import pathlib
import jnius_config
import numpy as np
import scipy.spatial.distance
import scipy.fft
import skimage.util
import skimage.util.dtype
import skimage.io
import skimage.exposure
import skimage.transform
import sklearn.linear_model
import networkx as nx
import tifffile
import matplotlib.pyplot as plt
import matplotlib.cm as mcm
import matplotlib.patches as mpatches
import matplotlib.patheffects as mpatheffects
import zarr
from . import utils
from . import thumbnail
from . import transform
from . import __version__ as _version


if not jnius_config.vm_running:
    pkg_root = pathlib.Path(__file__).parent.resolve()
    bf_jar_path = pkg_root / 'jars' / 'loci_tools.jar'
    if not bf_jar_path.exists():
        raise RuntimeError("loci_tools.jar missing from distribution"
                           " (expected it at %s)" % bf_jar_path)
    jnius_config.add_classpath(str(bf_jar_path))
    # These settings constrain the memory used by BioFormats to near the minimum
    # possible working set without requiring the choice of a max heap size.
    jnius_config.add_options("-Xms10m", "-XX:+UseSerialGC")

try:
    # let's try to run without jnius
    import jnius

    JBoolean = jnius.autoclass('java.lang.Boolean')
    DebugTools = jnius.autoclass('loci.common.DebugTools')
    IFormatReader = jnius.autoclass('loci.formats.IFormatReader')
    MetadataRetrieve = jnius.autoclass('ome.xml.meta.MetadataRetrieve')
    ServiceFactory = jnius.autoclass('loci.common.services.ServiceFactory')
    OMEXMLService = jnius.autoclass('loci.formats.services.OMEXMLService')
    ChannelSeparator = jnius.autoclass('loci.formats.ChannelSeparator')
    DynamicMetadataOptions = jnius.autoclass('loci.formats.in.DynamicMetadataOptions')
    UNITS = jnius.autoclass('ome.units.UNITS')
    DebugTools.enableLogging("ERROR")
    IS_JNIUS = True
except Exception as e:
    print(e)
    IS_JNIUS = False

# TODO:
# - Write tables with summary information about alignments.


class Metadata(object):

    @property
    def _num_images(self):
        raise NotImplementedError

    @property
    def num_channels(self):
        raise NotImplementedError

    @property
    def pixel_size(self):
        raise NotImplementedError
    
    @property
    def pixel_dtype(self):
        raise NotImplementedError

    def tile_position(self, i):
        raise NotImplementedError

    def tile_size(self, i):
        raise NotImplementedError

    @property
    def grid_dimensions(self):
        pos = self.positions
        shape = np.array([len(set(pos[:, d])) for d in range(2)])
        if np.prod(shape) != self.num_images:
            raise ValueError("Series positions do not form a grid")
        return shape

    @property
    def num_images(self):
        return self._num_images

    @property
    def positions(self):
        if not hasattr(self, '_positions'):
            self._positions = np.vstack([
                self.tile_position(i) for i in range(self._num_images)
            ])
        return self._positions

    @property
    def size(self):
        if not hasattr(self, '_size'):
            s0 = self.tile_size(0)
            image_ids = range(1, self._num_images)
            if any(any(self.tile_size(i) != s0) for i in image_ids):
                raise ValueError("Image series must all have the same dimensions")
            self._size = s0
        return self._size

    @property
    def centers(self):
        return self.positions + self.size / 2

    @property
    def origin(self):
        return self.positions.min(axis=0)


class PlateMetadata(Metadata):

    def __init__(self):
        super(PlateMetadata, self).__init__()
        self.set_active_plate_well(None, None)

    @property
    def num_plates(self):
        raise NotImplementedError

    @property
    def num_wells(self):
        raise NotImplementedError

    @property
    def plate_well_series(self):
        raise NotImplementedError

    def plate_name(self, i):
        raise NotImplementedError

    def well_name(self, plate, i):
        raise NotImplementedError

    def set_active_plate_well(self, plate, well):
        if (plate is None) ^ (well is None):
            raise ValueError("plate and well must be both set or both None")
        self.active_plate = plate
        self.active_well = well

    @property
    def active_series(self):
        if self.active_plate is None:
            return range(self._num_images)
        else:
            return self.plate_well_series[self.active_plate][self.active_well]

    @property
    def plate_names(self):
        if not hasattr(self, '_plate_names'):
            self._plate_names = [
                self.plate_name(i) for i in range(self.num_plates)
            ]
        return self._plate_names

    @property
    def well_names(self):
        if not hasattr(self, '_well_names'):
            self._well_names = [
                [self.well_name(p, i) for i in range(num_plate_wells)]
                for p, num_plate_wells in enumerate(self.num_wells)
            ]
        return self._well_names

    @Metadata.num_images.getter
    def num_images(self):
        return len(self.active_series)

    @Metadata.positions.getter
    def positions(self):
        return Metadata.positions.fget(self)[self.active_series]

    # FIXME Metadata.grid_dimensions should be overriden here or removed.


class Reader(object):

    def read(self, series, c):
        raise NotImplementedError


class PlateReader(Reader):
    # No API here, just a way to signal that a subclass's metadata class
    # inherits from PlateMetadata. This is probably a sign that the
    # architectural split between Metadata and Reader should be reconsidered.
    pass


class BioformatsMetadata(PlateMetadata):

    _pixel_dtypes = {
        'uint8': np.dtype(np.uint8),
        'uint16': np.dtype(np.uint16),
    }

    _ome_dtypes = {v: k for k, v in _pixel_dtypes.items()}

    def __init__(self, path):
        super(BioformatsMetadata, self).__init__()
        self.path = path
        self._init_metadata()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_reader'], state['_metadata'], state['_omexml_root']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._init_metadata()

    def _init_metadata(self):

        factory = ServiceFactory()
        service = jnius.cast(OMEXMLService, factory.getInstance(OMEXMLService))
        metadata = service.createOMEXMLMetadata()
        self._reader = ChannelSeparator()
        self._reader.setMetadataStore(metadata)
        self._reader.setFlattenedResolutions(False)
        # For multi-scene .CZI files, we need raw tiles instead of the
        # auto-stitched mosaic and we don't want labels or overview images
        options = DynamicMetadataOptions()
        options.setBoolean('zeissczi.autostitch', JBoolean(False))
        options.setBoolean('zeissczi.attachments', JBoolean(False))
        self._reader.setMetadataOptions(options)
        self._reader.setId(self.path)

        xml_content = service.getOMEXML(metadata)
        self._metadata = jnius.cast(MetadataRetrieve, metadata)
        self._omexml_root = xml.etree.ElementTree.fromstring(xml_content)
        self.format_name = self._reader.getFormat()

    @property
    def _num_images(self):
        count = self._metadata.imageCount
        # Skip final overview slide in Metamorph Slide Scan data if present.
        if (self.format_name == 'Metamorph STK'
            and 'overview' in self._metadata.getImageName(count - 1).lower()):
            count -= 1
        return count

    @property
    def num_channels(self):
        return self._metadata.getChannelCount(0)

    @property
    def num_plates(self):
        return self._metadata.getPlateCount()

    @property
    def num_wells(self):
        return [self._metadata.getWellCount(i) for i in range(self.num_plates)]

    @property
    def plate_well_series(self):
        if not hasattr(self, '_plate_well_series'):
            # FIXME Store slice objects to save resources where possible.
            self._plate_well_series = [
                [
                    [
                        self._metadata.getWellSampleIndex(p, w, s).value
                        for s in range(self._metadata.getWellSampleCount(p, w))
                    ]
                    for w in range(num_wells)
                ]
                for p, num_wells in enumerate(self.num_wells)
            ]
        return self._plate_well_series

    @property
    def pixel_size(self):
        values = []
        for dim in ('Y', 'X'):
            method = getattr(self._metadata, 'getPixelsPhysicalSize%s' % dim)
            v_units = method(0)
            if v_units is None:
                warn_data(
                    "Pixel size undefined; falling back to 1.0 \u03BCm."
                )
                value = 1.0
            else:
                value = v_units.value(UNITS.MICROMETER).doubleValue()
            values.append(value)
        values = tuple(values)
        if not np.isclose(values[0], values[1], rtol=1e-4, atol=0):
            raise Exception(f"Can't handle non-square pixels {values[::-1]}")
        if values[0] != values[1]:
            warn_data(
                f"Pixel size is slightly non-square {values[::-1]}. Using"
                f" {values[1]} for both dimensions."
            )
        return values[1]

    @property
    def pixel_dtype(self):
        return self._pixel_dtypes[self._metadata.getPixelsType(0).value]

    def plate_name(self, i):
        return self._metadata.getPlateName(i)

    @property
    def well_naming(self):
        if not hasattr(self, '_well_naming'):
            _well_naming = []
            for p in range(self.num_plates):
                row_nc = self._metadata.getPlateRowNamingConvention(p)
                column_nc = self._metadata.getPlateColumnNamingConvention(p)
                if row_nc is not None:
                    row_nc = row_nc.value
                else:
                    row_nc = 'letter'
                if column_nc is not None:
                    column_nc = column_nc.value
                else:
                    column_nc = 'number'
                if row_nc not in ('letter', 'number') or column_nc != 'number':
                    raise RuntimeError(
                        "Can't handle well naming convention row={} column={}"
                        .format(row_nc, column_nc)
                    )
                _well_naming.append([row_nc, column_nc])
            self._well_naming = _well_naming
        return self._well_naming

    def well_name(self, plate, i):
        row = self._metadata.getWellRow(plate, i).value
        column = self._metadata.getWellColumn(plate, i).value
        row_nc, column_nc = self.well_naming[plate]
        # FIXME Support formatting with 384/1536-well plates.
        assert row_nc in ('letter', 'number')
        assert column_nc == 'number'
        if row_nc == 'number':
            row_fmt = '{:02}'.format(row + 1)
        else:
            row_fmt = chr(ord('A') + row)
        column_fmt = '{:02}'.format(column + 1)
        return row_fmt + column_fmt

    def tile_position(self, i):
        planeCount = self._metadata.getPlaneCount(i)
        values = []
        for dim in ('Y', 'X'):
            method = getattr(self._metadata, 'getPlanePosition%s' % dim)
            # FIXME verify all planes have the same X,Y position.
            if planeCount > 0:
                # Returns None if planePositionX/Y not defined.
                v_units = method(i, 0)
            else:
                # Simple file formats don't have planes at all.
                v_units = None
            if v_units is None:
                warn_data(
                    "Stage coordinates undefined; falling back to (0, 0)."
                )
                values = [0.0, 0.0]
                break
            else:
                v = v_units.value(UNITS.MICROMETER)
                if v is None:
                    # Conversion failed, which usually happens when the unit is
                    # "reference frame". Proceed as if it's actually microns but
                    # emit a warning.
                    warn_data(
                        "Stage coordinates' measurement unit is undefined;"
                        " assuming \u03BCm."
                    )
                    v = v_units.value()
                value = v.doubleValue()
            values.append(value)
        position_microns = np.array(values, dtype=float)
        # Invert Y so that stage position coordinates and image pixel
        # coordinates are aligned (most formats seem to work this way).
        position_microns *= [-1, 1]
        position_pixels = position_microns / self.pixel_size
        return position_pixels

    def tile_size(self, i):
        values = []
        for dim in ('Y', 'X'):
            method = getattr(self._metadata, 'getPixelsSize%s' % dim)
            v = method(i).value
            values.append(v)
        return np.array(values, dtype=int)


import numpy as np

class NumpyReader(Reader):
    def __init__(self, array, metadata):
        self.array = array
        self.metadata = metadata

    def read(self, series, c):
        return self.array[series, c, :, :]

class NumpyMetadata(Metadata):
    def __init__(self, array, positions, pixel_size):
        self._positions = positions # list of (x, y) positions
        self._array = array # numpy array
        self.__num_images, self.__num_channels, self.__height, self.__width = array.shape
        self.__pixel_size = pixel_size  # You can set this to the actual pixel size

    @property
    def _num_images(self):
        return self.__num_images

    @property
    def num_channels(self):
        return self.__num_channels

    @property
    def pixel_size(self):
        return self.__pixel_size

    @property
    def pixel_dtype(self):
        return self._array.dtype

    def tile_position(self, i):
        # Example positions, replace with actual positions
        positions = np.array([[i * self.__width, j * self.__height] for i in range(int(np.sqrt(self.__num_images))) for j in range(int(np.sqrt(self.__num_images)))])
        return positions[i]

    def tile_size(self, i):
        return np.array([self.__height, self.__width])


class BioformatsReader(PlateReader):

    def __init__(self, path, plate=None, well=None):
        self.path = path
        self.metadata = BioformatsMetadata(self.path)
        self.metadata.set_active_plate_well(plate, well)

    def read(self, series, c):
        self.metadata._reader.setSeries(self.metadata.active_series[series])
        index = self.metadata._reader.getIndex(0, c, 0)
        byte_array = self.metadata._reader.openBytes(index)
        dtype = self.metadata.pixel_dtype
        shape = self.metadata.tile_size(series)
        img = np.frombuffer(byte_array.tostring(), dtype=dtype).reshape(shape)
        return img


class BarrelCorrectionReader(Reader):
    """Wraps a reader to correct barrel/pincushion image distortion."""

    def __init__(self, reader, k):
        self.reader = reader
        self.k = k

    @property
    def metadata(self):
        return self.reader.metadata

    def read(self, series, c):
        img = self.reader.read(series, c)
        img = transform.barrel_correction(img, self.k)
        img = utils.dtype_convert(img, self.metadata.pixel_dtype)
        return img


class CachingReader(Reader):
    """Wraps a reader to provide tile image caching."""

    def __init__(self, reader, channel):
        self.reader = reader
        self.channel = channel
        self._cache = {}

    @property
    def metadata(self):
        return self.reader.metadata

    def read(self, series, c):
        if c == self.channel and series in self._cache:
            img = self._cache[series]
        else:
            img = self.reader.read(series, c)
        if c == self.channel and series not in self._cache:
            self._cache[series] = img
        return img


# TileStatistics = collections.namedtuple(
#     'TileStatistics',
#     'scan tile x_original y_original x y shift_x shift_y error'
# )


@property
def neighbors_graph(aligner):
    """Return graph of neighboring (overlapping) tiles.

    Tiles are considered neighbors if the 'city block' distance between them
    is less than the largest tile dimension.

    """
    # FIXME: This should properly test for overlap, possibly via
    # intersection of bounding rectangles.
    if not hasattr(aligner, '_neighbors_graph'):
        pdist = scipy.spatial.distance.pdist(aligner.metadata.positions,
                                             metric='cityblock')
        sp = scipy.spatial.distance.squareform(pdist)
        max_distance = aligner.metadata.size.max() + 1
        edges = zip(*np.nonzero((sp > 0) & (sp < max_distance)))
        graph = nx.from_edgelist(edges)
        graph.add_nodes_from(range(aligner.metadata.num_images))
        aligner._neighbors_graph = graph
    return aligner._neighbors_graph


class EdgeAligner(object):

    def __init__(
        self, reader, channel=0, max_shift=15, alpha=0.01, max_error=None,
        randomize=False, filter_sigma=0.0, do_make_thumbnail=True, verbose=False
    ):
        self.channel = channel
        self.reader = CachingReader(reader, self.channel)
        self.verbose = verbose
        # Unit is micrometers.
        self.max_shift = max_shift
        self.max_shift_pixels = self.max_shift / self.metadata.pixel_size
        self.alpha = alpha
        self.max_error = max_error
        self.randomize = randomize
        self.filter_sigma = filter_sigma
        self.do_make_thumbnail = do_make_thumbnail
        self._cache = {}
        self.errors_negative_sampled = np.empty(0)

    neighbors_graph = neighbors_graph

    def run(self):
        self.make_thumbnail()
        self.check_overlaps()
        self.compute_threshold()
        self.register_all()
        self.build_spanning_tree()
        self.calculate_positions()
        self.fit_model()

    def make_thumbnail(self):
        if not self.do_make_thumbnail:
            return
        self.reader.thumbnail_scale = thumbnail.calculate_scale(self.reader)
        self.reader.thumbnail = thumbnail.make_thumbnail(
            self.reader, channel=self.channel, scale=self.reader.thumbnail_scale
        )

    def check_overlaps(self):
        # This might be better addressed by removing the +1 from the
        # neighbors_graph max_distance calculation and ensuring the graph is
        # fully connected.
        pos = self.metadata.positions
        overlaps = np.array([
            self.metadata.size - abs(pos[t1] - pos[t2])
            for t1, t2 in self.neighbors_graph.edges
        ])
        failures = np.any(overlaps < 1, axis=1) if len(overlaps) else []
        if len(failures) and all(failures):
            warn_data("No tiles overlap, attempting alignment anyway.")
        elif any(failures):
            warn_data("Some neighboring tiles have zero overlap.")

    def compute_threshold(self):
        if self.max_error is not None:
            if self.verbose:
                print("    using explicit error threshold")
            return
        # Compute error threshold for rejecting aligments. We generate a
        # distribution of error scores for many known non-overlapping image
        # regions and take a certain percentile as the maximum allowable error.
        # The percentile becomes our accepted false-positive ratio.
        edges = self.neighbors_graph.edges
        num_tiles = self.metadata.num_images
        # If not enough tiles overlap to matter, skip this whole thing.
        if len(edges) <= 1:
            self.max_error = np.inf
            return
        widths = np.array([
            self.intersection(t1, t2).shape.min()
            for t1, t2 in edges
        ])
        w = widths.max()
        max_offset = self.metadata.size[0] - w
        # Number of possible pairs minus number of actual neighbor pairs.
        num_distant_pairs = num_tiles * (num_tiles - 1) // 2 - len(edges)
        # Reduce permutation count for small datasets -- there are fewer
        # possible truly distinct strips with fewer tiles. The calculation here
        # is just a heuristic, not rigorously derived.
        n = 1000 if num_distant_pairs > 8 else (num_distant_pairs + 1) * 10
        pairs = np.empty((n, 2), dtype=int)
        offsets = np.empty((n, 2), dtype=int)
        # Generate n random non-overlapping image strips. Strips are always
        # horizontal, across the entire image width.
        max_tries = 100
        if self.randomize is False:
            random_state = np.random.RandomState(0)
        else:
            random_state = np.random.RandomState()
        for i in range(n):
            # Limit tries to avoid infinite loop in pathological cases.
            for current_try in range(max_tries):
                t1, t2 = random_state.randint(self.metadata.num_images, size=2)
                o1, o2 = random_state.randint(max_offset, size=2)
                # Check for non-overlapping strips and abort the retry loop.
                if t1 != t2 and (t1, t2) not in edges:
                    # Different, non-neighboring tiles -- always OK.
                    break
                elif t1 == t2 and abs(o1 - o2) > w:
                    # Same tile OK if strips don't overlap within the image.
                    break
                elif (t1, t2) in edges:
                    # Neighbors OK if either strip is entirely outside the
                    # expected overlap region (based on nominal positions).
                    its = self.intersection(t1, t2, np.repeat(w, 2))
                    ioff1, ioff2 = its.offsets[:, 0]
                    if (
                        its.shape[0] > its.shape[1]
                        or o1 < ioff1 - w or o1 > ioff1 + w
                        or o2 < ioff2 - w or o2 > ioff2 + w
                    ):
                        break
            else:
                # Retries exhausted. This should be very rare.
                warn_data(
                    "Could not find non-overlapping strips in {max_tries} tries"
                )
            pairs[i] = t1, t2
            offsets[i] = o1, o2
        errors = np.empty(n)
        for i, ((t1, t2), (offset1, offset2)) in enumerate(zip(pairs, offsets)):
            if self.verbose and (i % 10 == 9 or i == n - 1):
                sys.stdout.write(
                    '\r    quantifying alignment error %d/%d' % (i + 1, n)
                )
                sys.stdout.flush()
            img1 = self.reader.read(t1, self.channel)[offset1:offset1+w, :]
            img2 = self.reader.read(t2, self.channel)[offset2:offset2+w, :]
            _, errors[i] = utils.register(img1, img2, self.filter_sigma, upsample=1)
        if self.verbose:
            print()
        self.errors_negative_sampled = errors
        self.max_error = np.percentile(errors, self.alpha * 100)

    def register_all(self):
        n = self.neighbors_graph.size()
        for i, (t1, t2) in enumerate(self.neighbors_graph.edges, 1):
            if self.verbose:
                sys.stdout.write('\r    aligning edge %d/%d' % (i, n))
                sys.stdout.flush()
            self.register_pair(t1, t2)
        if self.verbose:
            print()
        self.all_errors = np.array([x[1] for x in self._cache.values()])
        # Set error values above the threshold to infinity.
        for k, v in self._cache.items():
            if v[1] > self.max_error or any(np.abs(v[0]) > self.max_shift_pixels):
                self._cache[k] = (v[0], np.inf)

    def build_spanning_tree(self):
        # Note that this may be disconnected, so it's technically a forest.
        g = nx.Graph()
        g.add_nodes_from(self.neighbors_graph)
        g.add_weighted_edges_from(
            (t1, t2, error)
            for (t1, t2), (_, error) in self._cache.items()
            if np.isfinite(error)
        )
        spanning_tree = nx.Graph()
        spanning_tree.add_nodes_from(g)
        for c in nx.connected_components(g):
            cc = g.subgraph(c)
            center = nx.center(cc)[0]
            paths = nx.single_source_dijkstra_path(cc, center).values()
            for path in paths:
                nx.add_path(spanning_tree, path)
        self.spanning_tree = spanning_tree

    def calculate_positions(self):
        shifts = {}
        for c in nx.connected_components(self.spanning_tree):
            cc = self.spanning_tree.subgraph(c)
            center = nx.center(cc)[0]
            shifts[center] = np.array([0, 0])
            for edge in nx.traversal.bfs_edges(cc, center):
                source, dest = edge
                if source not in shifts:
                    source, dest = dest, source
                shift = self.register_pair(source, dest)[0]
                shifts[dest] = shifts[source] + shift
        if shifts:
            self.shifts = np.array([s for _, s in sorted(shifts.items())])
            self.positions = self.metadata.positions + self.shifts
        else:
            # TODO: fill in shifts and positions with 0x2 arrays
            raise NotImplementedError("No images")

    def fit_model(self):
        components = sorted(
            nx.connected_components(self.spanning_tree),
            key=len, reverse=True
        )
        # Fit LR model on positions of largest connected component.
        cc0 = list(components[0])
        self.lr = sklearn.linear_model.LinearRegression()
        self.lr.fit(self.metadata.positions[cc0], self.positions[cc0])
        # Fix up degenerate transform matrix. This happens when the spanning
        # tree is completely edgeless or cc0's metadata positions fall in a
        # straight line. In this case we fall back to the identity transform.
        if np.linalg.det(self.lr.coef_) < 1e-3:
            # FIXME We should probably exit here, not just warn. We may provide
            # an option to force it anyway.
            warn_data(
                "Could not align enough edges, proceeding anyway with original"
                " stage positions."
            )
            self.lr.coef_ = np.diag(np.ones(2))
            self.lr.intercept_ = np.zeros(2)
        # Adjust position of remaining components so their centroids match
        # the predictions of the model.
        for cc in components[1:]:
            nodes = list(cc)
            centroid_m = np.mean(self.metadata.positions[nodes], axis=0)
            centroid_f = np.mean(self.positions[nodes], axis=0)
            shift = self.lr.predict([centroid_m])[0] - centroid_f
            self.positions[nodes] += shift
        # Adjust positions and model intercept to put origin at 0,0.
        self.origin = self.positions.min(axis=0)
        self.positions -= self.origin
        self.lr.intercept_ -= self.origin
        self.centers = self.positions + self.metadata.size / 2


    def register_pair(self, t1, t2):
        """Return relative shift between images and the alignment error."""
        key = tuple(sorted((t1, t2)))
        try:
            shift, error = self._cache[key]
        except KeyError:
            # We test a series of increasing overlap window sizes to help avoid
            # missing alignments when the stage position error is large relative
            # to the tile overlap. Simply using a large overlap in all cases
            # limits the maximum achievable correlation thus increasing the
            # error metric, leading to worse overall results. The window size
            # starts at the nominal size and doubles until it's at least 10% of
            # the tile size. If the nominal overlap is already 10% or greater,
            # we only use that one size.
            smin = self.intersection(key[0], key[1]).shape
            smax = np.round(self.metadata.size * 0.1)
            sizes = [smin]
            while any(sizes[-1] < smax):
                sizes.append(sizes[-1] * 2)
            results = [self._register(key[0], key[1], s) for s in sizes]
            # Use the shift from the window size that gave the lowest error.
            shift, _ = min(results, key=lambda r: r[1])
            # Extract the images from the nominal overlap window but with the
            # shift applied to the second tile's position, and compute the error
            # metric on these images. This should be even lower than the error
            # computed above.
            _, o1, o2 = self.overlap(key[0], key[1], shift=shift)
            error = utils.nccw(o1, o2, self.filter_sigma)
            self._cache[key] = (shift, error)
        if t1 > t2:
            shift = -shift
        # Return copy of shift to prevent corruption of cached values.
        return shift.copy(), error

    def _register(self, t1, t2, min_size=0):
        its, img1, img2 = self.overlap(t1, t2, min_size)
        # Account for padding, flipping the sign depending on the direction
        # between the tiles.
        p1, p2 = self.metadata.positions[[t1, t2]]
        sx = 1 if p1[1] >= p2[1] else -1
        sy = 1 if p1[0] >= p2[0] else -1
        padding = its.padding * [sy, sx]
        shift, error = utils.register(img1, img2, self.filter_sigma)
        shift += padding
        return shift, error

    def intersection(self, t1, t2, min_size=0, shift=None):
        corners1 = self.metadata.positions[[t1, t2]]
        if shift is not None:
            corners1[1] += shift
        corners2 = corners1 + self.metadata.size
        return Intersection(corners1, corners2, min_size)

    def crop(self, tile, offset, shape):
        img = self.reader.read(series=tile, c=self.channel)
        return utils.crop(img, offset, shape)

    def overlap(self, t1, t2, min_size=0, shift=None):
        its = self.intersection(t1, t2, min_size, shift)
        img1 = self.crop(t1, its.offsets[0], its.shape)
        img2 = self.crop(t2, its.offsets[1], its.shape)
        return its, img1, img2

    @property
    def best_edge(self):
        ordered_keys = sorted(self._cache, key=lambda k: self._cache[k][1])
        return ordered_keys[0]

    @property
    def metadata(self):
        return self.reader.metadata

    @property
    def mosaic_shape(self):
        upper_corners = self.positions + self.metadata.size
        max_dimensions = upper_corners.max(axis=0)
        return tuple(map(int, np.ceil(max_dimensions)))

    def debug(self, t1, t2, min_size=0):
        shift, _ = self._register(t1, t2, min_size)
        its, o1, o2 = self.overlap(t1, t2, min_size)
        w1 = utils.window(utils.whiten(o1, self.filter_sigma))
        w2 = utils.window(utils.whiten(o2, self.filter_sigma))
        corr = scipy.fft.fftshift(np.abs(scipy.fft.ifft2(
            scipy.fft.fft2(w1) * scipy.fft.fft2(w2).conj()
        )))
        corr /= (np.linalg.norm(w1) * np.linalg.norm(w2))
        stack = np.vstack
        rows, cols = 3, 1
        if corr.shape[0] > corr.shape[1]:
            stack = np.hstack
            rows, cols = cols, rows
        plt.figure()
        plt.subplot(rows, cols, 1)
        plt.imshow(stack([o1, o2]))
        ax = plt.subplot(rows, cols, 2)
        ax.set_xticks([])
        ax.set_yticks([])
        plt.imshow(stack([w1, w2]).real)
        ax = plt.subplot(rows, cols, 3)
        ax.set_xticks([])
        ax.set_yticks([])
        plt.imshow(corr, vmin=np.exp(-10))
        cbar = plt.colorbar()
        cbar.ax.yaxis.set_major_locator(
            plt.FixedLocator(cbar.mappable.get_clim())
        )
        cbar.ax.yaxis.set_major_formatter(
            plt.FuncFormatter(lambda x, pos: "{:.2f}".format(-np.log(x)))
        )
        origin = np.array(corr.shape) // 2
        plt.plot(origin[1], origin[0], 'r+')
        # FIXME This is wrong when t1 > t2.
        shift += origin + its.padding
        plt.plot(shift[1], shift[0], 'rx')
        plt.tight_layout()


class LayerAligner(object):

    def __init__(self, reader, reference_aligner, channel=None, max_shift=15,
                 filter_sigma=0.0, verbose=False):
        self.reader = reader
        self.reference_aligner = reference_aligner
        if channel is None:
            channel = reference_aligner.channel
        self.channel = channel
        # Unit is micrometers.
        self.max_shift = max_shift
        self.max_shift_pixels = self.max_shift / self.metadata.pixel_size
        self.filter_sigma = filter_sigma
        self.verbose = verbose
        # FIXME Still a bit muddled here on the use of metadata positions vs.
        # corrected positions from the reference aligner. We probably want to
        # use metadata positions to find the cycle-to-cycle tile
        # correspondences, but the corrected positions for computing our
        # corrected positions.

    neighbors_graph = neighbors_graph

    def run(self):
        self.make_thumbnail()
        self.coarse_align()
        self.register_all()
        self.calculate_positions()

    def make_thumbnail(self):
        self.reader.thumbnail = thumbnail.make_thumbnail(
            self.reader,
            channel=self.channel,
            scale=self.reference_aligner.reader.thumbnail_scale,
        )

    def coarse_align(self):
        self.cycle_offset = thumbnail.calculate_cycle_offset(
            self.reference_aligner.reader,
            self.reader,
            scale=self.reference_aligner.reader.thumbnail_scale,
        )
        self.corrected_nominal_positions = self.metadata.positions + self.cycle_offset
        reference_positions = self.reference_aligner.metadata.positions
        dist = scipy.spatial.distance.cdist(reference_positions,
                                            self.corrected_nominal_positions)
        self.reference_idx = np.argmin(dist, 0)
        self.reference_positions = reference_positions[self.reference_idx]
        self.reference_aligner_positions = self.reference_aligner.positions[self.reference_idx]

    def register_all(self):
        n = self.metadata.num_images
        self.shifts = np.empty((n, 2))
        self.errors = np.empty(n)
        for i in range(n):
            if self.verbose:
                sys.stdout.write("\r    aligning tile %d/%d" % (i + 1, n))
                sys.stdout.flush()
            shift, error = self.register(i)
            self.shifts[i] = shift
            self.errors[i] = error
        if self.verbose:
            print()

    def calculate_positions(self):
        self.positions = (
            self.corrected_nominal_positions
            + self.shifts
            + self.reference_aligner_positions
            - self.reference_positions
        )
        self.constrain_positions()
        self.centers = self.positions + self.metadata.size / 2

    def constrain_positions(self):
        position_diffs = np.absolute(
            self.positions - self.reference_aligner_positions
        )
        # Round the diffs to one decimal point because the subpixel shifts are
        # calculated by 10x upsampling and thus only accurate to that level.
        position_diffs = np.rint(position_diffs * 10) / 10
        # Discard camera background registration which will shift target
        # positions to reference aligner positions, due to strong
        # self-correlation of the sensor dark current pattern which dominates in
        # low-signal images.
        discard = (position_diffs == 0).all(axis=1)
        # Discard any tile registration that error is infinite
        discard |= np.isinf(self.errors)
        # Take the median of registered shifts to determine the offset
        # (translation) from the reference image to this one.
        if discard.all():
            offset = 0
        else:
            offset = np.nan_to_num(np.median(self.shifts[~discard], axis=0))
        # Here we assume the fitted linear model from the reference image is
        # still appropriate, apart from the extra offset we just computed.
        predictions = self.reference_aligner.lr.predict(self.corrected_nominal_positions)
        # Discard any tile registration that's too far from the linear model,
        # replacing it with the relevant model prediction.
        distance = np.linalg.norm(self.positions - predictions - offset, axis=1)
        max_dist = self.max_shift_pixels
        extremes = distance > max_dist
        # Recalculate the mean shift, also ignoring the extreme values.
        discard |= extremes
        self.discard = discard
        if discard.all():
            self.offset = 0
        else:
            self.offset = np.nan_to_num(np.mean(self.shifts[~discard], axis=0))
        # Fill in discarded shifts from the predictions.
        self.positions[discard] = predictions[discard] + self.offset

    def register(self, t):
        """Return relative shift between images and the alignment error."""
        its, ref_img, img = self.overlap(t)
        if np.any(np.array(its.shape) == 0):
            return (0, 0), np.inf
        shift, error = utils.register(ref_img, img, self.filter_sigma)
        # Add back in the fractional position difference that overlap() loses.
        shift = tuple(shift + its.offset_diff_frac)
        # We don't use padding and thus can skip the math to account for it.
        assert (its.padding == 0).all(), "Unexpected non-zero padding"
        return shift, error

    def intersection(self, t):
        corners1 = np.vstack([self.reference_positions[t],
                              self.corrected_nominal_positions[t]])
        corners2 = corners1 + self.reader.metadata.size
        its = Intersection(corners1, corners2)
        its.shape = its.shape // 32 * 32
        return its

    def overlap(self, t):
        its = self.intersection(t)
        ref_t = self.reference_idx[t]
        img1 = self.reference_aligner.reader.read(
            series=ref_t, c=self.reference_aligner.channel
        )
        img2 = self.reader.read(series=t, c=self.channel)
        # crop rounds the offsets to the nearest pixel, so the returned images
        # are not located at these precise locations.
        # Intersection.offset_diff_frac contains the difference.
        ov1 = utils.crop(img1, its.offsets[0], its.shape)
        ov2 = utils.crop(img2, its.offsets[1], its.shape)
        return its, ov1, ov2

    @property
    def metadata(self):
        return self.reader.metadata

    def debug(self, t):
        shift, _ = self.register(t)
        its, o1, o2 = self.overlap(t)
        w1 = utils.window(utils.whiten(o1, self.filter_sigma))
        w2 = utils.window(utils.whiten(o2, self.filter_sigma))
        corr = scipy.fft.fftshift(np.abs(scipy.fft.ifft2(
            scipy.fft.fft2(w1) * scipy.fft.fft2(w2).conj()
        )))
        plt.figure()
        plt.subplot(1, 3, 1)
        plt.imshow(np.vstack([o1, o2]))
        ax = plt.subplot(1, 3, 2)
        ax.set_xticks([])
        ax.set_yticks([])
        plt.imshow(np.vstack([w1, w2]).real)
        ax = plt.subplot(1, 3, 3)
        ax.set_xticks([])
        ax.set_yticks([])
        plt.imshow(corr)
        origin = np.array(corr.shape) // 2
        plt.plot(origin[1], origin[0], 'r+')
        shift += origin - its.offset_diff_frac
        plt.plot(shift[1], shift[0], 'rx')
        plt.tight_layout(0, 0, 0)


class Intersection(object):

    def __init__(self, corners1, corners2, min_size=0):
        if np.isscalar(min_size):
            min_size = np.repeat(min_size, 2)
        self._calculate(corners1, corners2, min_size)

    def _calculate(self, corners1, corners2, min_size):
        max_shape = (corners2 - corners1).max(axis=0)
        min_size = min_size.clip(1, max_shape)
        position = corners1.max(axis=0)
        initial_shape = np.floor(corners2.min(axis=0) - position).astype(int)
        clipped_shape = np.maximum(initial_shape, min_size)
        self.shape = np.ceil(clipped_shape).astype(int)
        self.padding = self.shape - initial_shape
        self.offsets = np.maximum(position - corners1 - self.padding, 0)
        offset_diff = self.offsets[1] - self.offsets[0]
        self.offset_diff_frac = offset_diff - offset_diff.round()

    def __repr__(self):
        s = 'shape: {0.shape}\npadding: {0.padding}\noffsets:\n{0.offsets}'
        return s.format(self)


class Mosaic(object):

    def __init__(
        self, aligner, shape, channels=None, ffp_path=None, dfp_path=None,
        flip_mosaic_x=False, flip_mosaic_y=False, barrel_correction=None,
        verbose=False
    ):
        self.aligner = aligner
        self.shape = tuple(shape)
        self.channels = self._sanitize_channels(channels)
        self.flip_mosaic_x = flip_mosaic_x
        self.flip_mosaic_y = flip_mosaic_y
        self.barrel_correction = barrel_correction
        self.dtype = aligner.metadata.pixel_dtype
        self._load_correction_profiles(dfp_path, ffp_path)
        self.verbose = verbose

    def _sanitize_channels(self, channels):
        all_channels = range(self.aligner.metadata.num_channels)
        if channels is None:
            channels = all_channels
        invalid_channels = sorted(set(channels) - set(all_channels))
        if invalid_channels:
            raise ValueError("invalid channels: %s" % invalid_channels)
        return channels

    def _load_single_profile(self, path, num_channels, img_size, profile_type):
        """Load, normalize, and validate illumination profile.

        Parameters
        ----------
        path : str
            Path to the image being loaded.
        num_channels : int
            Expected number of channels in the profile image.
        img_size : tuple
            Shape of a 2D image in (row, column).
        profile_type : str
            Type of profile, only accepts 'dark' and 'flat'.

        Returns
        ----------
        ndarray
            Image as numpy array in the (channel, row, column) arrangement.
            If ``path`` is ``None``, return an array in (channel, 1, 1) shape.
            The values in the array are 0 and 1 for dark- and flat-field profile, respectively.
        """
        assert profile_type in ('dark', 'flat'), "profile_type must be either 'dark' or 'flat'."
        if path is None:
            profile_shape = (num_channels, 1, 1)
            return (
                np.zeros(profile_shape)
                    if profile_type == 'dark'
                    else np.ones(profile_shape)
            )

        expected_ndim = 2 if num_channels == 1 else 3
        profile = skimage.io.imread(path)
        if profile.ndim != expected_ndim:
            raise ValueError(
                'Expect dimensionality is {} for {}-field profile but {} has {} dimensions.'.format(
                    expected_ndim, profile_type, path, profile.ndim
                )
            )

        profile = np.atleast_3d(profile)
        # skimage.io.imread convert images with 3 and 4 channels into (Y, X, C) shape,
        # but as (C, Y, X) for images with other channel numbers. We normalize
        # image-shape to (C, Y, X) regardless of the number of channels in the image.
        if num_channels in (1, 3, 4):
            profile = np.moveaxis(profile, 2, 0)
        if profile.shape != (num_channels,) + img_size:
            raise ValueError(
                '{}-field profile shape {} does not match target image shape {}.'.format(
                    profile_type.capitalize(), profile.shape, img_size
                )
            )

        if self.barrel_correction:
            cval = 0 if profile_type == "dark" else 1
            for cimg in profile:
                cimg[:] = transform.barrel_correction(
                    cimg, self.barrel_correction, cval=cval
                )

        return profile

    def _load_correction_profiles(self, dfp_path, ffp_path):
        if dfp_path is None and ffp_path is None:
            self.do_correction = False
        else:
            num_channels = self.aligner.metadata.num_channels
            img_size = tuple(self.aligner.metadata.size)
            self.dfp = self._load_single_profile(dfp_path, num_channels, img_size, 'dark')
            self.ffp = self._load_single_profile(ffp_path, num_channels, img_size, 'flat')

            # FIXME This assumes integer dtypes. Do we need to support floats?
            self.dfp /= np.iinfo(self.dtype).max
            self.do_correction = True

    def assemble_channel(self, channel, out=None):
        num_tiles = len(self.aligner.positions)
        if out is None:
            out = np.zeros(self.shape, self.dtype)
        else:
            if out.shape != self.shape:
                raise ValueError(
                    f"out array shape {out.shape} does not match Mosaic"
                    f" shape {self.shape}"
                )
        for si, position in enumerate(self.aligner.positions):
            if self.verbose:
                sys.stdout.write(f"\r        merging tile {si + 1}/{num_tiles}")
                sys.stdout.flush()
            img = self.aligner.reader.read(c=channel, series=si)
            img = self.correct_illumination(img, channel)
            utils.paste(out, img, position, func=utils.pastefunc_blend)
        # Memory-conserving axis flips.
        if self.flip_mosaic_x:
            for i in range(len(out)):
                out[i] = out[i, ::-1]
        if self.flip_mosaic_y:
            for i in range(len(out) // 2):
                out[[i, -i-1]] = out[[-i-1, i]]
        if self.verbose:
            print()
        return out

    def correct_illumination(self, img, channel):
        if self.do_correction:
            img = skimage.util.img_as_float(img, force_copy=True)
            img -= self.dfp[channel, ...]
            img /= self.ffp[channel, ...]
            img.clip(0, 1, out=img)
        return img


class PyramidWriter:

    def __init__(
        self, mosaics, path, scale=2, tile_size=1024, peak_size=1024, verbose=False
    ):
        if any(m.shape != mosaics[0].shape for m in mosaics[1:]):
            raise ValueError("mosaics must all have the same shape")
        if tile_size % 16 != 0:
            raise ValueError("tile_size must be a multiple of 16")
        self.mosaics = mosaics
        self.path = path
        self.scale = scale
        self.tile_size = tile_size
        self.peak_size = peak_size
        self.verbose = verbose

    @property
    def ref_mosaic(self):
        "The reference mosaic."
        return self.mosaics[0]

    @property
    def base_shape(self):
        "Shape of the base level."
        return self.ref_mosaic.shape

    @property
    def num_channels(self):
        return sum(len(m.channels) for m in self.mosaics)

    @property
    def num_levels(self):
        "Number of levels."
        factor = max(self.base_shape) / self.peak_size
        return max(math.ceil(math.log(factor, self.scale)) + 1, 1)

    @property
    def level_shapes(self):
        "Shape of all levels."
        factors = self.scale ** np.arange(self.num_levels)
        shapes = np.ceil(np.array(self.base_shape) / factors[:, None])
        return [tuple(map(int, s)) for s in shapes]

    @property
    def level_full_shapes(self):
        "Shape of all levels, including channel dimension."
        return [(self.num_channels, *shape) for shape in self.level_shapes]

    @property
    def tile_shapes(self):
        "Tile shape of all levels."
        level_shapes = np.array(self.level_shapes)
        # The last level where we want to use the standard square tile size.
        tip_level = np.argmax(np.all(level_shapes < self.tile_size, axis=1))
        tile_shapes = [
            (self.tile_size, self.tile_size) if i <= tip_level else None
            for i in range(len(level_shapes))
        ]
        return tile_shapes

    def base_tiles(self):
        h, w = self.base_shape
        th, tw = self.tile_shapes[0]
        for mi, mosaic in enumerate(self.mosaics):
            if self.verbose:
                print(f"Cycle {mi}:")
            for ci, channel in enumerate(mosaic.channels):
                if self.verbose:
                    print(f"    Channel {channel}:")
                img = mosaic.assemble_channel(channel)
                for y in range(0, h, th):
                    for x in range(0, w, tw):
                        # Returning a copy makes the array contiguous, avoiding
                        # a severely unoptimized code path in ndarray.tofile.
                        yield img[y:y+th, x:x+tw].copy()
                # Allow img to be freed immediately to avoid keeping it in
                # memory while the next loop iteration calls assemble_channel.
                img = None

    def subres_tiles(self, level):
        assert level >= 1
        num_channels, h, w = self.level_full_shapes[level]
        tshape = self.tile_shapes[level] or (h, w)
        tiff = tifffile.TiffFile(self.path)
        zimg = zarr.open(tiff.aszarr(series=0, level=level-1, squeeze=False))
        for c in range(num_channels):
            if self.verbose:
                sys.stdout.write(
                    f"\r        processing channel {c + 1}/{num_channels}"
                )
                sys.stdout.flush()
            th = tshape[0] * self.scale
            tw = tshape[1] * self.scale
            for y in range(0, zimg.shape[1], th):
                for x in range(0, zimg.shape[2], tw):
                    a = zimg[c, y:y+th, x:x+tw, 0]
                    a = skimage.transform.downscale_local_mean(
                        a, (self.scale, self.scale)
                    )
                    if np.issubdtype(zimg.dtype, np.integer):
                        a = np.around(a)
                    a = a.astype(zimg.dtype)
                    yield a

    def run(self):
        dtype = self.ref_mosaic.aligner.metadata.pixel_dtype
        pixel_size = self.ref_mosaic.aligner.metadata.pixel_size
        resolution_cm = 10000 / pixel_size
        software = f"Ashlar v{_version}"
        metadata = {
            "Creator": software,
            "Pixels": {
                "PhysicalSizeX": pixel_size, "PhysicalSizeXUnit": "\u00b5m",
                "PhysicalSizeY": pixel_size, "PhysicalSizeYUnit": "\u00b5m"
            },
        }
        with tifffile.TiffWriter(self.path, ome=True, bigtiff=True) as tiff:
            tiff.write(
                data=self.base_tiles(),
                metadata=metadata,
                software=software.encode("utf-8"),
                shape=self.level_full_shapes[0],
                subifds=int(self.num_levels - 1),
                dtype=dtype,
                tile=self.tile_shapes[0],
                resolution=(resolution_cm, resolution_cm, "centimeter"),
                # FIXME Propagate this from input files (especially RGB).
                photometric="minisblack",
                compression="adobe_deflate",
                predictor=True,
            )
            if self.verbose:
                print("Generating pyramid")
            for level, (shape, tile_shape) in enumerate(
                zip(self.level_full_shapes[1:], self.tile_shapes[1:]), 1
            ):
                if self.verbose:
                    print(f"    Level {level} ({shape[2]} x {shape[1]})")
                tiff.write(
                    data=self.subres_tiles(level),
                    shape=shape,
                    subfiletype=1,
                    dtype=dtype,
                    tile=tile_shape,
                    compression="adobe_deflate",
                    predictor=True,
                )
                if self.verbose:
                    print()


class TiffListWriter:

    def __init__(self, mosaics, path_format, verbose=False):
        if any(m.shape != mosaics[0].shape for m in mosaics[1:]):
            raise ValueError("mosaics must all have the same shape")
        self.mosaics = mosaics
        self.path_format = path_format
        self.verbose = verbose

    def run(self):
        pixel_size = self.mosaics[0].aligner.metadata.pixel_size
        resolution_cm = 10000 / pixel_size
        software = f"Ashlar v{_version}"
        used_paths = set()
        for mi, mosaic in enumerate(self.mosaics):
            if self.verbose:
                print(f"Cycle {mi}:")
            for ci, channel in enumerate(mosaic.channels):
                if self.verbose:
                    print(f"    Channel {channel}:")
                path = self.path_format.format(cycle=mi, channel=channel)
                # FIXME: We would ideally report this and exit immediately
                # rather than warn after all the alignment has been done, but
                # that will require some refactoring to open each input file
                # right away to discover channel counts.
                if path in used_paths:
                    warn_data(
                        f"Output path collision -- overwriting {path} (please"
                        " include both {cycle} and {channel} placeholders in"
                        " the output path, or use the .ome.tif extension to"
                        " write all channels to one file)"
                    )
                used_paths.add(path)
                img = mosaic.assemble_channel(channel)
                with tifffile.TiffWriter(path, bigtiff=True) as tiff:
                    tiff.write(
                        data=img,
                        software=software.encode("utf-8"),
                        resolution=(resolution_cm, resolution_cm, "centimeter"),
                        # FIXME Propagate this from input files (especially RGB).
                        photometric="minisblack",
                    )
                # Allow img to be freed.
                img = None


class Warning(UserWarning):
    """Warnings about ashlar operation."""
    pass


class DataWarning(Warning):
    """Warnings about the content of user-provided image data."""
    pass


def warn_data(message):
    warnings.warn(message, DataWarning)


def plot_edge_shifts(aligner, img=None, bounds=True, im_kwargs=None):
    if im_kwargs is None:
        im_kwargs = {}
    fig = plt.figure()
    ax = plt.gca()
    draw_mosaic_image(ax, aligner, img, **im_kwargs)
    h, w = aligner.reader.metadata.size
    if bounds:
        # Bounding boxes denoting new tile positions.
        for xy in np.fliplr(aligner.positions):
            rect = mpatches.Rectangle(xy, w, h, color='black', fill=False,
                                      lw=0.5)
            ax.add_patch(rect)
    # Compute per-edge relative shifts from tile positions.
    edges = np.array(list(aligner.spanning_tree.edges))
    dist = aligner.metadata.positions - aligner.positions
    shifts = dist[edges[:, 0]] - dist[edges[:, 1]]
    shift_distances = np.linalg.norm(shifts, axis=1)
    # Spanning tree with nodes at new tile positions, edges colored by shift
    # distance (brighter = farther).
    nx.draw(
        aligner.spanning_tree, ax=ax, with_labels=True,
        pos=np.fliplr(aligner.centers), edge_color=shift_distances,
        edge_cmap=plt.get_cmap('Blues_r'), width=2, node_size=100, font_size=6
    )
    fig.set_facecolor('black')


def plot_edge_quality(
    aligner, img=None, show_tree=True, pos='metadata', im_kwargs=None, nx_kwargs=None
):
    if pos == 'metadata':
        centers = aligner.metadata.centers - aligner.metadata.origin
    elif pos == 'aligner':
        centers = aligner.centers
    else:
        raise ValueError("pos must be either 'metadata' or 'aligner'")
    if im_kwargs is None:
        im_kwargs = {}
    if nx_kwargs is None:
        nx_kwargs = {}
    final_nx_kwargs = dict(width=2, node_size=100, font_size=6)
    final_nx_kwargs.update(nx_kwargs)
    if show_tree:
        nrows, ncols = 1, 2
        if aligner.mosaic_shape[1] * 2 / aligner.mosaic_shape[0] > 2 * 4 / 3:
            nrows, ncols = ncols, nrows
    else:
        nrows, ncols = 1, 1
    fig = plt.figure()
    ax = plt.subplot(nrows, ncols, 1)
    draw_mosaic_image(ax, aligner, img, **im_kwargs)
    error = np.array([aligner._cache[tuple(sorted(e))][1]
                      for e in aligner.neighbors_graph.edges])
    # Manually center and scale data to 0-1, except infinity which is set to -1.
    # This lets us use the purple-green diverging color map to color the graph
    # edges and cause the "infinity" edges to disappear into the background
    # (which is itself purple).
    infs = error == np.inf
    error[infs] = -1
    if not infs.all():
        error_f = error[~infs]
        emin = np.min(error_f)
        emax = np.max(error_f)
        if emin == emax:
            # Always true when there's only one edge. Otherwise it's unlikely
            # but theoretically possible.
            erange = 1
        else:
            erange = emax - emin
        error[~infs] = (error_f - emin) / erange
    # Neighbor graph colored by edge alignment quality (brighter = better).
    nx.draw(
        aligner.neighbors_graph, ax=ax, with_labels=True,
        pos=np.fliplr(centers), edge_color=error, edge_vmin=-1, edge_vmax=1,
        edge_cmap=plt.get_cmap('PRGn'), **final_nx_kwargs
    )
    if show_tree:
        ax = plt.subplot(nrows, ncols, 2)
        draw_mosaic_image(ax, aligner, img, **im_kwargs)
        # Spanning tree with nodes at original tile positions.
        nx.draw(
            aligner.spanning_tree, ax=ax, with_labels=True,
            pos=np.fliplr(centers), edge_color='royalblue',
            **final_nx_kwargs
        )
    fig.set_facecolor('black')


def plot_edge_scatter(aligner, annotate=True):
    import seaborn as sns
    xdata = np.clip(aligner.all_errors, 0, 10)
    ydata = np.clip(
        [np.linalg.norm(v[0]) for v in aligner._cache.values()], 0.01, np.inf
    )
    pdata = np.clip(aligner.errors_negative_sampled, 0, 10)
    g = sns.JointGrid(x=xdata, y=ydata)
    g.plot_joint(sns.scatterplot, alpha=0.5)
    _, xbins = np.histogram(np.hstack([xdata, pdata]), bins=40)
    sns.distplot(
        xdata, ax=g.ax_marg_x, kde=False, bins=xbins, norm_hist=True
    )
    sns.distplot(
        pdata, ax=g.ax_marg_x, kde=False, bins=xbins, norm_hist=True,
        hist_kws=dict(histtype='step')
    )
    g.ax_joint.axvline(aligner.max_error, c='k', ls=':')
    g.ax_joint.axhline(aligner.max_shift_pixels, c='k', ls=':')
    g.ax_joint.set_yscale('log')
    g.set_axis_labels('error', 'shift')
    if annotate:
        for pair, x, y in zip(aligner.neighbors_graph.edges, xdata, ydata):
            g.ax_joint.annotate(str(pair), (x, y), alpha=0.1)
    plt.tight_layout()


def plot_layer_shifts(aligner, img=None, im_kwargs=None):
    if im_kwargs is None:
        im_kwargs = {}
    fig = plt.figure()
    ax = plt.gca()
    draw_mosaic_image(ax, aligner, img, **im_kwargs)
    h, w = aligner.metadata.size
    # Bounding boxes denoting new tile positions.
    for xy in np.fliplr(aligner.positions):
        rect = mpatches.Rectangle(xy, w, h, color='black', fill=False, lw=0.5)
        ax.add_patch(rect)
    # Neighbor graph with edges hidden, i.e. just show nodes.
    nx.draw(
        aligner.neighbors_graph, ax=ax, with_labels=True,
        pos=np.fliplr(aligner.centers), edge_color='none',
        node_size=100, font_size=6
    )
    fig.set_facecolor('black')


def plot_layer_quality(
    aligner, img=None, scale=1.0, artist='patches', annotate=True, im_kwargs=None
):
    if im_kwargs is None:
        im_kwargs = {}
    fig = plt.figure()
    ax = plt.gca()
    draw_mosaic_image(ax, aligner, img, **im_kwargs)

    h, w = aligner.metadata.size
    positions, centers, shifts = aligner.positions, aligner.centers, aligner.shifts

    if scale != 1.0:
        h, w, positions, centers, shifts = [
            scale * i for i in [h, w, positions, centers, shifts]
        ]

    # Bounding boxes denoting new tile positions.
    color_index = skimage.exposure.rescale_intensity(
        aligner.errors, out_range=np.uint8
    ).astype(np.uint8)
    color_map = mcm.magma_r
    for xy, c_idx in zip(np.fliplr(positions), color_index):
        rect = mpatches.Rectangle(
            xy, w, h, color=color_map(c_idx), fill=False, lw=0.5
        )
        ax.add_patch(rect)
    
    # Annotate tile numbering.
    if annotate:
        for idx, (x, y) in enumerate(np.fliplr(positions)):
            text = plt.annotate(str(idx), (x+0.1*w, y+0.9*h), alpha=0.7)
            # Add outline to text for better contrast in different background color.
            text_outline = mpatheffects.Stroke(linewidth=1, foreground='#AAA')
            text.set_path_effects(
                [text_outline, mpatheffects.Normal()]
            )

    if artist == 'quiver':
        ax.quiver(
            *centers.T[::-1], *shifts.T[::-1], aligner.discard,
            units='dots', width=2, scale=1, scale_units='xy', angles='xy',
            cmap='Greys'
        )
    if artist == 'patches':
        for xy, dxy, is_discarded in zip(
            np.fliplr(centers), np.fliplr(shifts), aligner.discard
        ):
            arrow = mpatches.FancyArrowPatch(
                xy, np.array(xy) + np.array(dxy), 
                arrowstyle='->', color='0' if is_discarded else '1',
                mutation_scale=8,
                )
            ax.add_patch(arrow)
    ax.axis('off')


def draw_mosaic_image(ax, aligner, img, **kwargs):
    if img is None:
        img = [[0]]
    h, w = aligner.mosaic_shape
    ax.imshow(img, extent=(-0.5, w-0.5, h-0.5, -0.5), **kwargs)



import re
import itertools
import pathlib
import numpy as np
import skimage.io
from . import reg


# Classes for reading datasets consisting of TIFF files with a naming pattern.
# The pattern must include an integer series number, and optionally a channel
# name or number, and well name.
#
# This code is experimental and probably still has a lot of rough edges.


def format_to_regex(s):
    # Translate a restricted subset of the "format" pattern language to
    # a matching regex with named capture.
    s = s.replace('.', '\.')
    s = s.replace('(', '\(')
    s = s.replace(')', '\)')
    regex = re.sub(r'{([^:}]+):?([^}]*)}', f2r_repl, s)
    return regex

def f2r_repl(m):
    r = '(?P<' + m.group(1) + '>.'
    if re.match(r'^\d+$', m.group(2)):
        r += '{' + m.group(2) + '}'
    else:
        r += '*?'
    r += ')'
    return r

rgb_suffixes = "jpg jpeg gif png".split()

def _read(path, channel=None):
    suffix = path.suffix[1:]
    path = str(path)
    try:
        if suffix in rgb_suffixes and channel is not None:
            # RGB image types with shape (M, N, 3) or (M, N, 4) that don't
            # support reading a single plane in isolation. Ideally we wouldn't
            # read the whole file just to extract one channel!
            img = skimage.io.imread(path)[..., channel]
        else:
            kwargs = {}
            if channel is not None:
                kwargs["key"] = channel
            img = skimage.io.imread(path, **kwargs)
    except:
        reader = reg.BioformatsReader(path)
        if channel is None:
            channels = range(reader.metadata.num_channels)
            img = np.stack([reader.read(0, c) for c in channels])
        else:
            img = reader.read(0, channel)
    # Undo skimage's "helpful" reordering of 3- and 4-channel images.
    if (
        img.ndim == 3
        and img.shape[2] in (3, 4)
        and img.shape[0] not in (3, 4)
    ):
        img = np.moveaxis(img, 2, 0)
    return img


class FileSeriesMetadata(reg.PlateMetadata):

    def __init__(
        self, path, pattern, overlap, width, height, layout, direction,
        pixel_size,
    ):
        # The pattern argument uses the Python Format String syntax with
        # required "series" and optionally "channel" fields. A width
        # specification with leading zeros must be used for any fields that are
        # zero-padded. The pattern is used both to parse the filenames upon
        # initialization as well as to synthesize filenames when reading images.
        # Example pattern: 'img_s{series}_w{channel}.tif'
        super(FileSeriesMetadata, self).__init__()
        self.path = pathlib.Path(path)
        self.pattern = pattern
        self.overlap = overlap
        self.width = width
        self.height = height
        self.layout = layout
        self.direction = direction
        self._pixel_size = pixel_size
        self._enumerate_tiles()

    def _enumerate_tiles(self):
        regex = format_to_regex(self.pattern)
        wells = set()
        series = set()
        channels = set()
        n = 0
        self.filename_components = {}
        for p in self.path.iterdir():
            match = re.match(regex, p.name)
            if match:
                gd = match.groupdict()
                w = gd.get('well')
                s = int(gd['series'])
                c = gd.get('channel')
                wells.add(w)
                series.add(s)
                channels.add(c)
                self.filename_components[w, s, c] = gd
                n += 1
        if len(self.filename_components) != len(wells) * len(series) * len(channels):
            raise Exception("Missing images detected")
        # Build sorted list of (well, series) tuples for all wells.
        self.all_series = sorted(set(
            k[:2] for k in self.filename_components.keys()
        ))
        self.well_map = dict(enumerate(sorted(wells)))
        self._actual_num_images = len(series) * len(wells)
        self.channel_map = dict(enumerate(sorted(channels)))
        path = self.path / self.filename(0, 0)
        img = _read(path)
        if img.ndim not in (2, 3):
            raise Exception(f"Image must have 2 or 3 dimensions: {path}")
        self._tile_size = np.array(img.shape[-2:])
        self._dtype = img.dtype
        self.multi_channel_tiles = False
        # Handle multi-channel tiles (pattern must not include channel).
        if len(self.channel_map) == 1 and img.ndim == 3:
            self.channel_map = {c: None for c in range(img.shape[0])}
            self.multi_channel_tiles = True
        self._num_channels = len(self.channel_map)

    @property
    def _num_images(self):
        return self._actual_num_images

    @property
    def num_channels(self):
        return self._num_channels

    # This only supports a single plate at the moment.
    @property
    def num_plates(self):
        return 1

    @property
    def num_wells(self):
        return [len(self.well_map)]

    @property
    def plate_well_series(self):
        return [
            list([a[0] for a in v] for k, v in itertools.groupby(enumerate(self.all_series), key=lambda x: x[1][0]))
        ]

    def plate_name(self, i):
        assert i == 0, "Plate index out of range"
        return "Plate_1"

    def well_name(self, plate, i):
        assert plate == 0, "Plate index out of range"
        return self.well_map[i]

    @property
    def pixel_size(self):
        return self._pixel_size

    @property
    def pixel_dtype(self):
        return self._dtype

    def tile_position(self, i):
        row, col = self.tile_rc(i)
        return [row, col] * self.tile_size(i) * (1 - self.overlap)

    def tile_size(self, i):
        return self._tile_size

    def tile_rc(self, i):
        if self.direction == "horizontal":
            row = i // self.width
            col = i % self.width
            if self.layout == "snake" and row % 2 == 1:
                col = self.width - 1 - col
        else:
            row = i % self.height
            col = i // self.height
            if self.layout == "snake" and col % 2 == 1:
                row = self.height - 1 - row
        return row, col

    def filename(self, series, c):
        well, series = self.all_series[self.active_series[series]]
        c = self.channel_map[c]
        components = self.filename_components[well, series, c]
        return self.pattern.format(**components)


class FileSeriesReader(reg.PlateReader):

    def __init__(
        self, path, pattern, overlap, width, height, layout="raster",
        direction="horizontal", pixel_size=1.0, plate=None, well=None
    ):
        # See FileSeriesMetadata for an explanation of the pattern syntax.
        if layout not in ("raster", "snake"):
            raise ValueError("layout must be 'raster' or 'snake'")
        if direction not in ("horizontal", "vertical"):
            raise ValueError("direction must be 'horizontal' or 'vertical'")
        self.path = pathlib.Path(path)
        self.pattern = pattern
        self.metadata = FileSeriesMetadata(
            self.path, self.pattern, overlap, width, height, layout, direction,
            pixel_size
        )
        self.metadata.set_active_plate_well(plate, well)

    def read(self, series, c):
        # TODO: Address tension between non-plate and plate-aware modes
        # here and in Metadata class.
        path = self.path / self.metadata.filename(series, c)
        channel = c if self.metadata.multi_channel_tiles else 0
        img = _read(path, channel)
        return img



import sys
import pathlib
import numpy as np
from . import utils
from skimage.transform import rescale
from skimage.registration import phase_cross_correlation
import tifffile


def calculate_scale(reader, default_scale=0.05, min_size=1000):
    """Return scaling factor for a thumbnail with a minimum size constraint."""
    positions = reader.metadata.positions - reader.metadata.origin
    full_shape = (positions + reader.metadata.size).max(axis=0)
    thumbnail_shape = full_shape * default_scale
    if any(thumbnail_shape >= min_size):
        scale = default_scale
    else:
        # Compute the scale needed to make the thumbnail min_size on the longest
        # side, but don't scale smaller images up.
        scale = min(min_size / max(full_shape), 1)
    return scale


def make_thumbnail(reader, channel=0, scale=0.05):
    positions = reader.metadata.positions - reader.metadata.origin
    full_shape = (positions + reader.metadata.size).max(axis=0)
    mshape = np.ceil(full_shape * scale).astype(int)
    mosaic = np.zeros(mshape, dtype=np.uint16)
    total = reader.metadata.num_images
    for i, pos_s in enumerate(positions * scale):
        sys.stdout.write("\r    assembling thumbnail %d/%d" % (i + 1, total))
        sys.stdout.flush()
        img = reader.read(c=channel, series=i)
        # We don't need anti-aliasing as long as the coarse features in the
        # images are bigger than the scale factor. This speeds up the rescaling
        # dramatically.
        img_s = rescale(img, scale, anti_aliasing=False)
        utils.paste(mosaic, img_s, pos_s, utils.pastefunc_blend)
    print()
    return mosaic


def calculate_image_offset(img1, img2, upsample_factor=1):
    ref = utils.window(utils.whiten(img1, 1))
    test = utils.window(utils.whiten(img2, 1))
    shift = phase_cross_correlation(
        ref,
        test,
        upsample_factor=upsample_factor,
        normalization=None,
        return_error=False,
    )
    return shift


def calculate_cycle_offset(reader1, reader2, scale=0.05):
    if not hasattr(reader1, 'thumbnail'):
        raise ValueError('reader1 does not have a thumbnail')
    if not hasattr(reader2, 'thumbnail'):
        raise ValueError('reader2 does not have a thumbnail')
    img1 = reader1.thumbnail
    img2 = reader2.thumbnail
    if img1.shape != img2.shape:
        padded_shape = np.array((img1.shape, img2.shape)).max(axis=0)
        padded_img1, padded_img2 = np.zeros(padded_shape), np.zeros(padded_shape)
        utils.paste(padded_img1, img1, [0, 0])
        utils.paste(padded_img2, img2, [0, 0])
        img1 = padded_img1
        img2 = padded_img2
    img_offset = calculate_image_offset(img1, img2, int(1/scale)) / scale
    img_offset -= (reader2.metadata.origin - reader1.metadata.origin)
    print(
        '\r    estimated cycle offset [y x] =',
        img_offset
    )
    return img_offset


def _save_as_tif(img, file_path, post_fix=''):
    input_path = pathlib.Path(file_path)
    if input_path.is_dir():
        raise RuntimeError("file_path must point to a file not a directory")
    filename = input_path.name.replace('.', post_fix + '.', 1)
    out_path = input_path.parent / filename
    out_path = out_path.with_suffix('.tif')
    tifffile.imwrite(out_path, img)



import napari
import numpy as np
import networkx as nx


_colormaps = ['red', 'green', 'blue', 'magenta', 'cyan', 'yellow']


def view_edges(
        aligner,
        viewer=None,
        tiles=None,
        vmin=None,
        vmax=None,
        gamma=1/2.2,
):
    """View EdgeAligner results in a Napari viewer using a checkerboard style.

    This "checkerboard" visualization colors tiles in pure red, green, and blue
    and applies additive blending to make good or bad alignment easy to spot at
    a per-pixel level. Pure yellow, cyan, and magenta objects indicate good
    alignment, whereas red, green, and blue fringes indicate misalignment.

    All tiles' contrast limits and gamma are linked so you can adjust those
    settings for all tiles by adjusting any one.

    Parameters
    ----------
    aligner: EdgeAligner
        The EdgeAligner to visualize.
    viewer: optional napari.Viewer
        If present, add layers to this viewer instead of creating a new one.
    tiles: optional List[int]
        If present, visualize only the given tiles instead of every one.
    vmin, vmax: optional numeric
        Initial contrast limits. Detected from data if not provided.
    gamma: optional numeric
        Initial gamma correction. Default is appropriate for linear data.

    Returns
    -------
    napari.viewer.Viewer

    """

    if tiles is None:
        tiles = range(aligner.metadata.num_images)
    else:
        if not set(tiles) <= set(range(aligner.metadata.num_images)):
            raise ValueError("tiles contains invalid tile index values")

    node_colors = nx.greedy_color(aligner.neighbors_graph)
    num_colors = max(node_colors.values()) + 1
    if num_colors > len(_colormaps):
        raise ValueError(f"neighbors_graph requires more than {len(_colormaps)} colors")

    dtype = aligner.metadata.pixel_dtype
    if np.issubdtype(dtype, np.integer):
        info = np.iinfo(dtype)
    elif np.issubdtype(dtype, np.floating):
        info = np.iinfo(dtype)
    else:
        raise ValueError(f"Can't display {dtype} image data")
    compute_vmin = vmin is None
    compute_vmax = vmax is None
    if compute_vmin:
        dmin = info.min
        vmin = np.inf
    else:
        dmin = vmin
    if compute_vmax:
        dmax = info.max
        vmax = -np.inf
    else:
        dmax = vmax

    # We wait until now, after we've passed all the argument validations, to
    # create the viewer window. It seems like bad form to throw up an empty
    # window only to then error out due to a bad arg value.
    if viewer is None:
        viewer = napari.Viewer()

    images = []

    # Link contrast sliders for all tiles.
    def contrast_limits_callback(event):
        source_image = event.source
        new_clims = source_image.contrast_limits
        for image in images:
            # this is critical to avoid a RecursionError
            if image._contrast_limits != new_clims:
                image.contrast_limits = new_clims

    # Link gamma sliders for all tiles.
    def gamma_callback(event):
        source_image = event.source
        new_gamma = source_image.gamma
        for image in images:
            # this is critical to avoid a RecursionError
            if image._gamma != new_gamma:
                image.gamma = new_gamma

    for i in tiles:
        data = aligner.reader.read(i, aligner.channel)
        image = viewer.add_image(
            data,
            name=str(i),
            translate=aligner.positions[i],
            colormap=_colormaps[node_colors[i]],
            contrast_limits=(dmin, dmax),
            gamma=gamma,
            blending='additive',
            interpolation='linear',
        )
        image.events.contrast_limits.connect(contrast_limits_callback)
        image.events.gamma.connect(gamma_callback)
        images.append(image)
        if compute_vmin:
            vmin = min(vmin, np.min(data))
        if compute_vmax:
            vmax = max(vmax, np.max(data))
    if compute_vmin or compute_vmax:
        for image in images:
            image.contrast_limits = (vmin, vmax)

    bounds_min = aligner.positions[tiles].min(axis=0)
    bounds_max = aligner.positions[tiles].max(axis=0) + aligner.metadata.size
    rect = tuple(bounds_min[::-1]) + tuple(bounds_max[::-1])
    viewer.events.reset_view(rect=rect)

    return viewer



import pathlib
from urllib.parse import unquote as urllib_unquote
import xml.etree.ElementTree
import numpy as np
import skimage.io
from . import reg


class ZenMetadata(reg.Metadata):

    def __init__(self, path):
        self.path = pathlib.Path(path).resolve()
        self.base_path = self.path.parent
        self._init_metadata()

    def _init_metadata(self):
        tree = xml.etree.ElementTree.parse(str(self.path))
        positions_map = {}
        tile_size = None
        self._num_channels = 0
        self.image_paths = {}
        for image in tree.findall('Image'):
            path = urllib_unquote(image.findtext('Filename'))
            bounds = image.find('Bounds')
            series = int(bounds.attrib['StartM'])
            channel = int(bounds.attrib['StartC'])
            start_x = int(bounds.attrib['StartX'])
            start_y = int(bounds.attrib['StartY'])
            size_x = int(bounds.attrib['SizeX'])
            size_y = int(bounds.attrib['SizeY'])
            position = start_y, start_x
            size = size_y, size_x
            if channel == 0:
                positions_map[series] = position
            else:
                if positions_map[series] != position:
                    print(
                        "Position for series %d, channel %d doesn't match"
                        " channel 0." % (series, channel)
                    )
            if tile_size is None:
                tile_size = size
            else:
                if tile_size != size:
                    raise Exception(
                        "Size for series %d, channel %d doesn't match"
                        " first image." % (series, channel)
                    )
            self.image_paths[series, channel] = path
            self._num_channels = max(self._num_channels, channel + 1)
        positions = [pos for series, pos in sorted(positions_map.items())]
        self._positions = np.array(positions, np.float64)
        self._tile_size = np.array(tile_size, np.int64)

    @property
    def _num_images(self):
        return len(self._positions)

    @property
    def num_channels(self):
        return self._num_channels

    @property
    def pixel_size(self):
        return 1.0

    @property
    def pixel_dtype(self):
        return np.dtype(np.uint16)

    def tile_size(self, i):
        return self._tile_size

    def image_path(self, series, c):
        return self.base_path / self.image_paths[series, c]


class ZenReader(reg.Reader):

    def __init__(self, path):
        self.metadata = ZenMetadata(path)
        self.path = pathlib.Path(path)

    def read(self, series, c):
        path = self.metadata.image_path(series, c)
        img = skimage.io.imread(str(path), key=0)
        return img



import os
from urllib.request import urlopen
import hashlib
from setuptools import setup, find_packages
from setuptools.command.develop import develop
from setuptools.command.sdist import sdist
from setuptools.command.build_py import build_py


from setuptools import setup, find_packages

# Version will be read from your package's __init__.py
# Make sure __version__ is defined in imswitch/__init__.py
def get_version():
    version_file = 'ashlarUC2/__init__.py'
    with open(version_file, 'r') as file:
        for line in file:
            if line.startswith('__version__'):
                # Strip the line to remove whitespaces and newline characters,
                # then split the string by '=' and take the second part,
                # which is the version number. Finally, strip the quotes and any
                # additional whitespace or newline characters around the version number.
                version = line.strip().split('=')[1].strip().strip('\'"')
                return version
    raise RuntimeError("Version konnte nicht gefunden werden.")


requires = [
    'numpy>=1.18.1',
    'pyjnius>=1.2.1',
    'matplotlib>=3.1.2',
    'networkx>=2.4',
    'scipy>=1.4.1',
    'scikit-image>=0.19.2,<0.20', # The v1.0 API promises breaking changes.
    'scikit-learn>=0.21.1',
    'tifffile>=2023.3.15',
    'zarr>=2.11.3',
    'blessed>=1.17' #'imagecodecs>=2021.6.8',
]

# get version from init file
VERSION =  get_version()
DESCRIPTION = ('Alignment by Simultaneous Harmonization of Layer/Adjacency '
               'Registration')
LONG_DESCRIPTION='''

ASHLAR: Alignment by Simultaneous Harmonization of Layer/Adjacency Registration

Ashlar implements efficient combined stitching and registration of multi-channel
image mosaics collected using the Tissue-CycIF microscopy protocol [1]_. Although
originally developed for CycIF, it may also be applicable to other tiled and/or
cyclic imaging approaches. The package offers both a command line script for the
most common use cases as well as an API for building more specialized tools.

.. [1] Tissue-CycIF is multi-round immunofluorescence microscopy on large fixed
   tissue samples. See https://doi.org/10.1101/151738 for details.

'''
AUTHOR = 'Jeremy Muhlich modified by Benedict Diederich'
AUTHOR_EMAIL = 'benedictdied@gmail.com'
LICENSE = 'MIT License'
HOMEPAGE = 'https://github.com/openuc2/ashlar'

LOCI_TOOLS_URL = 'https://downloads.openmicroscopy.org/bio-formats/6.3.1/artifacts/loci_tools.jar'
LOCI_TOOLS_SHA1 = 'bdf1a37b561fea02fd8d1c747bd34db3fc49667b'

def download_bioformats():
    print("Ensuring latest bioformats is present:")
    dist_root = os.path.abspath(os.path.dirname(__file__))
    jar_dir = os.path.join(dist_root, 'ashlarUC2', 'jars')
    lt_jar_path = os.path.join(jar_dir, 'loci_tools.jar')
    if not os.access(jar_dir, os.F_OK):
        os.mkdir(jar_dir)
    try:
        with open(lt_jar_path, 'rb') as f:
            existing_sha1 = hashlib.sha1(f.read()).hexdigest()
            if existing_sha1 == LOCI_TOOLS_SHA1:
                print("    Up to date!")
                return
    except IOError:
        pass
    print("    Downloading BioFormats from %s ..." % LOCI_TOOLS_URL)
    # FIXME add progress bar
    content = urlopen(LOCI_TOOLS_URL).read()
    content_sha1 = hashlib.sha1(content).hexdigest()
    with open(lt_jar_path, 'wb') as f:
        f.write(content)
    if content_sha1 != LOCI_TOOLS_SHA1:
        raise RuntimeError("loci_tools.jar hash mismatch")

# Define some distutils command subclasses for a few key commands to trigger
# downloading the BioFormats JAR before they run.

class PreDevelop(develop):
    def run(self):
        download_bioformats()
        develop.run(self)

class PreSdist(sdist):
    def run(self):
        download_bioformats()
        sdist.run(self)

class PreBuildPy(build_py):
    def run(self):
        download_bioformats()
        build_py.run(self)

cmdclass = {
    'develop': PreDevelop,
    'sdist': PreSdist,
    'build_py': PreBuildPy,
}

setup(
    name='ashlarUC2',
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/x-rst',
    packages=find_packages(),
    include_package_data=True,
    install_requires=requires,
    entry_points={
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: %s' % LICENSE,
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Visualization'
    ],
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    license=LICENSE,
    url=HOMEPAGE,
    download_url='%s/archive/v%s.tar.gz' % (HOMEPAGE, VERSION),
    keywords=['scripts', 'microscopy', 'registration', 'stitching'],
    zip_safe=False,
)



import warnings
import sys
import re
import argparse
import pathlib
import blessed
import os
import numpy as np

try:
    from .. import __version__ as VERSION
    from .. import reg
    from ..reg import PlateReader, BioformatsReader, NumpyReader, NumpyMetadata
    from ..filepattern import FilePatternReader
    from ..fileseries import FileSeriesReader
    from ..zen import ZenReader
except:
    from ashlarUC2 import __version__ as VERSION
    from ashlarUC2.reg import PlateReader, BioformatsReader, NumpyReader, NumpyMetadata
    from ashlarUC2.filepattern import FilePatternReader
    from ashlarUC2.fileseries import FileSeriesReader
    from ashlarUC2.zen import ZenReader


def main(argv=sys.argv):

    parser = argparse.ArgumentParser(
        description='Stitch and align multi-tile cyclic microscope images',
        formatter_class=HelpFormatter,
    )
    parser.add_argument(
        'filepaths', metavar='FILE', nargs='+',
        help='Image file(s) to be processed, one per cycle',
    )
    parser.add_argument(
        '-o', '--output', dest='output', default='ashlar_output.ome.tif',
        metavar='PATH',
        help=("Output file. If PATH ends in .ome.tif a pyramidal OME-TIFF will"
              " be written. If PATH ends in just .tif and includes {cycle} and"
              " {channel} placeholders, a series of single-channel plain TIFF"
              " files will be written. If PATH starts with a relative or"
              " absolute path to another directory, that directory must already"
              " exist."),
    )
    parser.add_argument(
        '-c', '--align-channel', dest='align_channel', type=int,
        default='0', metavar='CHANNEL',
        help=('Reference channel number for image alignment. Numbering starts'
              ' at 0.'),
    )
    parser.add_argument(
        '--flip-x', default=False, action='store_true',
        help='Flip tile positions left-to-right',
    )
    parser.add_argument(
        '--flip-y', default=False, action='store_true',
        help='Flip tile positions top-to-bottom',
    )
    parser.add_argument(
        '--flip-mosaic-x', default=False, action='store_true',
        help='Flip output image left-to-right',
    )
    parser.add_argument(
        '--flip-mosaic-y', default=False, action='store_true',
        help='Flip output image top-to-bottom',
    )
    parser.add_argument(
        '--output-channels', nargs='+', type=int, metavar='CHANNEL',
        help=('Output only specified channels for each cycle. Numbering starts'
              ' at 0. (default: all channels)'),
    )
    parser.add_argument(
        '-m', '--maximum-shift', type=float, default=15, metavar='SHIFT',
        help='Maximum allowed per-tile corrective shift in microns',
    )
    parser.add_argument(
        "--stitch-alpha", type=float, default=0.01, metavar="ALPHA",
        help="Significance level for permutation testing during alignment error"
        " quantification. Larger values include more tile pairs in the spanning"
        " tree at the cost of increased false positives."
    )
    parser.add_argument(
        "--maximum-error", type=float, default=None, help=argparse.SUPPRESS,
    )
    parser.add_argument(
        '--filter-sigma', type=float, default=0, metavar='SIGMA',
        help=('Filter images before alignment using a Gaussian kernel with s.d.'
              ' of SIGMA pixels (default: no filtering)'),
    )
    parser.add_argument(
        '-f', '--filename-format', dest='filename_format',
        default='cycle_{cycle}_channel_{channel}.tif', help=argparse.SUPPRESS,
    )
    parser.add_argument(
        '--pyramid', default=False, action='store_true', help=argparse.SUPPRESS
    )
    parser.add_argument(
        '--tile-size', type=int, default=1024, metavar='PIXELS',
        help='Pyramid tile size for OME-TIFF output',
    )
    parser.add_argument(
        '--ffp', metavar='FILE', nargs='+',
        help=("Perform flat field illumination correction using the given"
              " profile image. Specify one common file for all cycles or one"
              " file for every cycle. Channel counts must match input files."
              " (default: no flat field correction)"),
    )
    parser.add_argument(
        '--dfp', metavar='FILE', nargs='+',
        help=("Perform dark field illumination correction using the given"
              " profile image. Specify one common file for all cycles or one"
              " file for every cycle. Channel counts must match input files."
              " (default: no dark field correction)"),
    )
    parser.add_argument(
        "--barrel-correction", type=float, default=0, help=argparse.SUPPRESS
    )
    parser.add_argument(
        '--plates', default=False, action='store_true',
        help='Enable plate mode for HTS data',
    )
    parser.add_argument(
        '-q', '--quiet', dest='quiet', default=False,
        action='store_true', help='Suppress progress display',
    )
    parser.add_argument(
        '--version', action='version', version=f"ashlar {VERSION}"
    )
    args = parser.parse_args(argv[1:])

    configure_terminal()
    configure_warning_format()

    filepaths = args.filepaths

    output_path = pathlib.Path(args.output)
    op_tiff = bool(re.search(r"\.tiff?$", output_path.name, re.IGNORECASE))
    ff_default = args.filename_format == parser.get_default("filename_format")
    if op_tiff and ff_default:
        # Standard usage: -o includes a .tif filename, -f not included.
        args.filename_format = output_path.name
        output_path = output_path.parent
    else:
        # Old, deprecated usage: -o is a directory and/or -f was specified.
        if ff_default:
            warnings.warn(
                "The output path must include a filename with a .tif or .tiff"
                " suffix. Specifying only a directory path with -o/--output has"
                " been deprecated and will be disabled in a future version. See"
                " the -o documentation for details.",
                reg.Warning,
            )
        else:
            warnings.warn(
                "The -f/--filename-format argument has been deprecated and its"
                " functionality merged into the -o argument. See the -o"
                " documentation for details.",
                reg.Warning,
            )
        if op_tiff and not output_path.is_dir():
            # Checking is_dir() avoids erroring out in the strange but legal
            # situation where output_path is a DIRECTORY that ends in .tif !
            print_error(
                "Filename may be appended to the output path specified by"
                " -o/--output, or specified separately with"
                " -f/--filename-format, but not both at the same time"
            )
            return 1
        if not re.search(r"\.tiff?$", args.filename_format, re.IGNORECASE):
            print_error(
                f"Filename format does not end in .tif: {args.filename_format}"
            )
            return 1
    if not output_path.is_dir():
        print_error(
            "Output location does not exist or is not a directory:"
            f" {output_path}"
        )
        return 1
    if re.search(r"\.ome\.tiff?$", args.filename_format, re.IGNORECASE):
        args.pyramid = True

    if args.tile_size != parser.get_default("tile_size") and not args.pyramid:
        print_error("--tile-size can only be used with OME-TIFF output")
        return 1

    ffp_paths = args.ffp
    if ffp_paths:
        if len(ffp_paths) not in (0, 1, len(filepaths)):
            print_error(
                "Wrong number of flat-field profiles. Must be 1, or {}"
                " (number of input files)".format(len(filepaths))
            )
            return 1
        if len(ffp_paths) == 1:
            ffp_paths = ffp_paths * len(filepaths)

    dfp_paths = args.dfp
    if dfp_paths:
        if len(dfp_paths) not in (0, 1, len(filepaths)):
            print_error(
                "Wrong number of dark-field profiles. Must be 1, or {}"
                " (number of input files)".format(len(filepaths))
            )
            return 1
        if len(dfp_paths) == 1:
            dfp_paths = dfp_paths * len(filepaths)

    aligner_args = {}
    aligner_args['channel'] = args.align_channel
    aligner_args['verbose'] = not args.quiet
    aligner_args['max_shift'] = args.maximum_shift
    aligner_args["alpha"] = args.stitch_alpha
    aligner_args["max_error"] = args.maximum_error
    aligner_args['filter_sigma'] = args.filter_sigma

    mosaic_args = {}
    if args.output_channels:
        mosaic_args['channels'] = args.output_channels
    if args.pyramid:
        mosaic_args['tile_size'] = args.tile_size
    if args.quiet is False:
        mosaic_args['verbose'] = True
    mosaic_args['flip_mosaic_x'] = args.flip_mosaic_x
    mosaic_args['flip_mosaic_y'] = args.flip_mosaic_y

    try:
        if args.plates:
            return process_plates(
                filepaths, output_path, args.filename_format, args.flip_x,
                args.flip_y, ffp_paths, dfp_paths, args.barrel_correction,
                aligner_args, mosaic_args, args.pyramid, args.quiet
            )
        else:
            mosaic_path_format = str(output_path / args.filename_format)
            return process_single(
                filepaths, mosaic_path_format, args.flip_x, args.flip_y,
                ffp_paths, dfp_paths, args.barrel_correction, aligner_args,
                mosaic_args, args.pyramid, args.quiet
            )
    except ProcessingError as e:
        print_error(str(e))
        return 1


def process_single(
    filepaths, output_path_format, flip_x, flip_y, ffp_paths, dfp_paths,
    barrel_correction, aligner_args, mosaic_args, pyramid, quiet, plate_well=None, position_list=None, pixel_size=None
):
    
    mosaic_args = mosaic_args.copy()
    writer_args = {}
    if pyramid:
        writer_args["tile_size"] = mosaic_args.pop("tile_size", None)
    mosaics = []


    if isinstance(filepaths[0], np.ndarray):
        reader = build_numpy_reader(filepaths[0], position_list=position_list, pixel_size=pixel_size, plate_well=plate_well)
    else:
        if not quiet:
            print("Stitching and registering input images")
            print('Cycle 0:')
            print('    reading %s' % filepaths[0])
        reader = build_reader(filepaths[0], barrel_correction, plate_well=plate_well)
    process_axis_flip(reader, flip_x, flip_y)
    ea_args = aligner_args.copy()
    if len(filepaths) == 1:
        ea_args['do_make_thumbnail'] = False
    edge_aligner = reg.EdgeAligner(reader, **ea_args)
    edge_aligner.run()
    mshape = edge_aligner.mosaic_shape
    mosaic_args_final = mosaic_args.copy()
    if ffp_paths:
        mosaic_args_final['ffp_path'] = ffp_paths[0]
    if dfp_paths:
        mosaic_args_final['dfp_path'] = dfp_paths[0]
    mosaics.append(reg.Mosaic(edge_aligner, mshape, **mosaic_args_final))

    for cycle, filepath in enumerate(filepaths[1:], 1):
        if not quiet:
            print('Cycle %d:' % cycle)
            print('    reading %s' % filepath)
        if isinstance(filepath, np.ndarray):
            reader = build_numpy_reader(filepath, plate_well=plate_well)
        else:
            reader = build_reader(filepath, barrel_correction, plate_well=plate_well)
        process_axis_flip(reader, flip_x, flip_y)
        layer_aligner = reg.LayerAligner(reader, edge_aligner, **aligner_args)
        layer_aligner.run()
        mosaic_args_final = mosaic_args.copy()
        if ffp_paths:
            mosaic_args_final['ffp_path'] = ffp_paths[cycle]
        if dfp_paths:
            mosaic_args_final['dfp_path'] = dfp_paths[cycle]
        mosaics.append(reg.Mosaic(layer_aligner, mshape, **mosaic_args_final))

    # Disable reader caching to save memory during mosaicing and writing.
    edge_aligner.reader = edge_aligner.reader.reader

    if not quiet:
        print()
        print(f"Merging tiles and writing to {output_path_format}")
    writer_class = reg.PyramidWriter if pyramid else reg.TiffListWriter
    writer = writer_class(
        mosaics, output_path_format, verbose=not quiet, **writer_args
    )
    writer.run()

    return 0


def process_plates(
    filepaths, output_path, filename_format, flip_x, flip_y, ffp_paths,
    dfp_paths, barrel_correction, aligner_args, mosaic_args, pyramid, quiet
):

    temp_reader = build_reader(filepaths[0])
    metadata = temp_reader.metadata
    if metadata.num_plates == 0:
        # FIXME raise ProcessingError here instead?
        print("Dataset does not contain plate information.")
        return 1

    for p, plate_name in enumerate(metadata.plate_names):
        print("Plate {} ({})\n==========\n".format(p, plate_name))
        for w, well_name in enumerate(metadata.well_names[p]):
            print("Well {}\n-----".format(well_name))
            if len(metadata.plate_well_series[p][w]) > 0:
                plate_path = output_path / plate_name
                if pyramid:
                    out_file_path = plate_path / f"{well_name}_{filename_format}"
                else:
                    out_file_path = plate_path / well_name / filename_format
                out_file_path.parent.mkdir(parents=True, exist_ok=True)
                mosaic_path_format = str(out_file_path)
                process_single(
                    filepaths, mosaic_path_format, flip_x, flip_y,
                    ffp_paths, dfp_paths, barrel_correction, aligner_args,
                    mosaic_args, pyramid, quiet, plate_well=(p, w), position_list=position_list, pixel_size=pixel_size
                )
            else:
                print("Skipping -- No images found.")
            print()
        print()

    return 0


def process_axis_flip(reader, flip_x, flip_y):
    metadata = reader.metadata
    # Trigger lazy initialization.
    _ = metadata.positions
    sx = -1 if flip_x else 1
    sy = -1 if flip_y else 1
    metadata._positions *= [sy, sx]


readers = {
    'filepattern': FilePatternReader,
    'fileseries': FileSeriesReader,
    'bioformats': BioformatsReader,
    'zen': ZenReader,
    'numpy': NumpyReader,
}


def build_numpy_reader(array, position_list, pixel_size, plate_well=None):
    metadata = NumpyMetadata(array, position_list, pixel_size)
    return NumpyReader(array, metadata)

# This is a short-term hack to provide a way to specify alternate reader
# classes and pass specific args to them.
def build_reader(path, barrel_correction=0, plate_well=None):
    # Default to BioformatsReader if name not specified.
    reader_class = BioformatsReader
    kwargs = {}
    match = re.match(
        r'(?P<reader>\w+)\|(?P<path>.*?)(\|(?P<kwargs>.*))?$', path
    )
    if match:
        path = match.group('path')
        reader_name = match.group('reader')
        reader_class = readers.get(reader_name)
        if reader_class is None:
            raise ProcessingError("Unknown reader: {}".format(reader_name))
        kwargs.update(parse_kwargs_string(match.group('kwargs')))
    if plate_well is not None:
        if not issubclass(reader_class, PlateReader):
            raise ProcessingError(
                "The %s reader does not support plate/well processing"
                % reader_class.__name__
            )
        kwargs.update(plate=plate_well[0], well=plate_well[1])
    reader = reader_class(path, **kwargs)
    if barrel_correction != 0:
        reader = reg.BarrelCorrectionReader(reader, barrel_correction)
    return reader


def parse_kwargs_string(string):
    kwargs = {}
    if string is not None:
        for piece in string.split('|'):
            name, value = piece.split('=')
            # Optimistically parse as float.
            try:
                value = float(value)
            except ValueError:
                pass
            kwargs[name] = value
    return kwargs


def configure_terminal():
    global terminal
    terminal = blessed.Terminal()


def print_error(message):
    print(terminal.bright_red("ERROR:"), message)


def warning_formatter(message, category, filename, lineno, line=None):
    if issubclass(category, reg.Warning):
        return terminal.bright_yellow("WARNING:") + f" {message}\n"
    else:
        return _old_formatwarning(message, category, filename, lineno, line)


def configure_warning_format():
    global _old_formatwarning
    _old_formatwarning = warnings.formatwarning
    warnings.formatwarning = warning_formatter


class HelpFormatter(argparse.HelpFormatter):
    """Help message formatter which adds default values to argument help.

    Based on argparse.ArgumentDefaultsHelpFormatter.
    """

    def _get_help_string(self, action):
        help = action.help
        if isinstance(action, (argparse._HelpAction, argparse._VersionAction)):
            help = help.capitalize()
        elif (
            not isinstance(action, argparse._StoreTrueAction)
            and "%(default)" not in help
            and "(default:" not in help
            and action.default is not argparse.SUPPRESS
        ):
            defaulting_nargs = [argparse.OPTIONAL, argparse.ZERO_OR_MORE]
            if action.option_strings or action.nargs in defaulting_nargs:
                help += " (default: %(default)s)"
        return help
class ProcessingError(RuntimeError):
    pass

def process_images(
    filepaths,
    output='ashlar_output.ome.tif',
    align_channel=0,
    flip_x=False,
    flip_y=False,
    flip_mosaic_x=False,
    flip_mosaic_y=False,
    output_channels=None,
    maximum_shift=15,
    stitch_alpha=0.01,
    maximum_error=None,
    filter_sigma=0,
    filename_format='cycle_{cycle}_channel_{channel}.tif',
    pyramid=False,
    tile_size=1024,
    ffp=None,
    dfp=None,
    barrel_correction=0,
    plates=False,
    quiet=False, 
    position_list=None,
    pixel_size=0.5
):
    configure_terminal()
    configure_warning_format()
    
    # preapre positionlist if provided
    if position_list is not None:
        position_microns = np.fliplr(np.array(position_list, dtype=float)) # make it consistent with OMEReader (e.g. XY instead of YX)
        # Invert Y so that stage position coordinates and image pixel
        # coordinates are aligned (most formats seem to work this way).
        position_microns *= [-1, 1]
        position_list = position_microns / pixel_size

    output_path = pathlib.Path(output)
    op_tiff = bool(re.search(r"\.tiff?$", output_path.name, re.IGNORECASE))
    ff_default = filename_format == 'cycle_{cycle}_channel_{channel}.tif'
    if op_tiff and ff_default:
        # Standard usage: -o includes a .tif filename, -f not included.
        filename_format = output_path.name
        output_path = output_path.parent
    else:
        # Old, deprecated usage: -o is a directory and/or -f was specified.
        if ff_default:
            warnings.warn(
                "The output path must include a filename with a .tif or .tiff"
                " suffix. Specifying only a directory path with -o/--output has"
                " been deprecated and will be disabled in a future version. See"
                " the -o documentation for details.",
                reg.Warning,
            )
        else:
            warnings.warn(
                "The -f/--filename-format argument has been deprecated and its"
                " functionality merged into the -o argument. See the -o"
                " documentation for details.",
                reg.Warning,
            )
        if op_tiff and not output_path.is_dir():
            # Checking is_dir() avoids erroring out in the strange but legal
            # situation where output_path is a DIRECTORY that ends in .tif !
            print_error(
                "Filename may be appended to the output path specified by"
                " -o/--output, or specified separately with"
                " -f/--filename-format, but not both at the same time"
            )
            return 1
        if not re.search(r"\.tiff?$", filename_format, re.IGNORECASE):
            print_error(
                f"Filename format does not end in .tif: {filename_format}"
            )
            return 1
    if not output_path.is_dir():
        print_error(
            "Output location does not exist or is not a directory:"
            f" {output_path}"
        )
        return 1
    if re.search(r"\.ome\.tiff?$", filename_format, re.IGNORECASE):
        pyramid = True

    if tile_size != 1024 and not pyramid:
        print_error("--tile-size can only be used with OME-TIFF output")
        return 1

    ffp_paths = ffp
    if ffp_paths:
        if len(ffp_paths) not in (0, 1, len(filepaths)):
            print_error(
                "Wrong number of flat-field profiles. Must be 1, or {}"
                " (number of input files)".format(len(filepaths))
            )
            return 1
        if len(ffp_paths) == 1:
            ffp_paths = ffp_paths * len(filepaths)

    dfp_paths = dfp
    if dfp_paths:
        if len(dfp_paths) not in (0, 1, len(filepaths)):
            print_error(
                "Wrong number of dark-field profiles. Must be 1, or {}"
                " (number of input files)".format(len(filepaths))
            )
            return 1
        if len(dfp_paths) == 1:
            dfp_paths = dfp_paths * len(filepaths)

    aligner_args = {
        'channel': align_channel,
        'verbose': not quiet,
        'max_shift': maximum_shift,
        'alpha': stitch_alpha,
        'max_error': maximum_error,
        'filter_sigma': filter_sigma
    }

    mosaic_args = {
        'flip_mosaic_x': flip_mosaic_x,
        'flip_mosaic_y': flip_mosaic_y,
        'verbose': not quiet
    }
    if output_channels:
        mosaic_args['channels'] = output_channels
    if pyramid:
        mosaic_args['tile_size'] = tile_size

    try:
        if plates:
            return process_plates(
                filepaths, output_path, filename_format, flip_x,
                flip_y, ffp_paths, dfp_paths, barrel_correction,
                aligner_args, mosaic_args, pyramid, quiet, position_list=position_list, pixel_size=pixel_size
            )
        else:
            mosaic_path_format = str(output_path / filename_format)
            return process_single(
                filepaths, mosaic_path_format, flip_x, flip_y,
                ffp_paths, dfp_paths, barrel_correction, aligner_args,
                mosaic_args, pyramid, quiet, position_list=position_list, pixel_size=pixel_size
            )
    except ProcessingError as e:
        print_error(str(e))
        return 1

if __name__ == '__main__':
    import numpy as np

    # Create example numpy arrays
    num_images = 4
    num_channels = 3
    height, width = 256, 256
    arrays = [np.random.rand(num_images, num_channels, height, width) for _ in range(2)]

    # Process numpy arrays
    process_images(arrays, 'output_path_format', 0, False, False, False, False, None, 15, 0.01, None, 0, False, 1024, None, None, 0, False, False)
    #process_single(arrays, 'output_path_format', False, False, None, None, aligner_args, mosaic_args, False, False)

